from abc import ABC, abstractmethod, abstractproperty
from typing import Any, Sequence
import tkinter as tk

import numpy as np
import pandas as pd
from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.backend_bases import key_press_handler
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk  # type: ignore

from pstl.utls import constants as c
from pstl.diagnostics.probes.langmuir.single.classes import available_plasma_properties
from pstl.diagnostics.probes.langmuir.single.analysis.solver import SingleLangmuirProbeSolver


class Widgets:
    # Initialize the class with empty dictionaries for each widget name
    def __init__(self):
        self.buttons = {}
        self.canvases = {}
        self.checkbuttons = {}
        self.entries = {}
        self.frames = {}
        self.labels = {}
        self.labelframes = {}
        self.listboxes = {}
        self.menus = {}
        self.menubuttons = {}
        self.messages = {}
        self.optionmenus = {}
        self.panedwindows = {}
        self.radiobuttons = {}
        self.scales = {}
        self.scrollbars = {}
        self.texts = {}
        self.toplevels = {}


class WidgetOrganizer:
    def __init__(self) -> None:
        self._widgets = Widgets()

    @property
    def widget(self):
        return self._widgets


class MatplotlibCanvas(tk.Frame):
    def __init__(self, master=None, cnf=None, *args, fig: Figure | None = None, ax: Axes | None = None,  **kwargs):
        width = kwargs.pop("width", 8)
        height = kwargs.pop("height", 6)
        dpi = kwargs.pop("dpi", 100)
        super().__init__(master, cnf, *args, **kwargs)

        # add widgets
        self.widgets = Widgets()

        if fig is None:
            # initialize matplotlib figure and axes

            self.fig = Figure(figsize=(width, height), dpi=dpi)
            self.fig.set_tight_layout(True)
        else:
            self.fig = fig

        if ax is None:
            self.ax = self.fig.add_subplot(
                1, 1, 1)  # type: ignore
        else:
            self.ax = ax

        # A tk.DrawingArea.
        self.canvas = FigureCanvasTkAgg(self.fig, master=self)
        self.canvas.draw()

        # Pack Canvas
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)

    def set_fig_to_linear(self):
        fig = self.fig

        ax = self.ax
        scale = ax.get_yscale()
        if len(ax.lines) == 0:
            ax.set_yscale("linear")
        else:
            ydata = ax.lines[0].get_ydata()
            ydata_minimum = np.min(ydata)
            ydata_maximum = np.max(ydata)
            ylim = ax.get_ylim()

            ypad = 0.05*(np.max(ydata)-np.min(ydata))

            ylim = [ydata_minimum-ypad, ydata_maximum+ypad]

            ax.set_yscale("linear")

            ax.set_ylim(ylim[0], ylim[1])
        fig.canvas.draw()

    def set_fig_to_semilogy(self):
        fig = self.fig

        ax = self.ax
        scale = ax.get_yscale()

        if len(ax.lines) == 0:
            ax.set_yscale("log")
        else:
            ydata = ax.lines[0].get_ydata()

            # Set the y-axis limits with minimum positive value padding
            y_lim = ax.get_ylim()
            y_min = np.floor(np.log10(np.min(ydata[ydata > 0])))
            y_max = np.ceil(np.log10(y_lim[1]))

            ax.set_yscale("log")

            ax.set_ylim(10**(y_min), 10**y_max)
        fig.canvas.draw()


class MatplotlibCanvasWToolbar(MatplotlibCanvas):
    def __init__(self, master=None, cnf=None, *args, **kwargs):
        super().__init__(master, cnf, *args, **kwargs)
        # pack_toolbar=False will make it easier to use a layout manager later on.
        self.toolbar = NavigationToolbar2Tk(
            self.canvas, self, pack_toolbar=False)
        self.toolbar.update()

        # Pack Toolbar
        self.toolbar.pack(side=tk.TOP, fill=tk.X, expand=True)


class MatplotlibCanvasWToolbarSave(MatplotlibCanvasWToolbar):
    def __init__(self, master=None,
                 cnf=None, *args,
                 fig: Figure | None = None, ax: Axes | None = None,
                 saveas: str = "figure1.png",
                 **kwargs):
        super().__init__(master, cnf, *args, fig=fig, ax=ax, **kwargs)

        self.saveas = saveas

        # add save default button
        self.widgets.buttons['save'] = tk.Button(
            self, text="Save", command=self.save)
        self.widgets.buttons['save'].pack(
            side=tk.TOP, fill=tk.X, expand=True)

    def save(self, saveas: str | None = None, **kwargs):
        if saveas is None:
            saveas = self.saveas
        self.fig.savefig(saveas, **kwargs)


class LinearSemilogyCanvas(MatplotlibCanvasWToolbarSave):
    def __init__(self, master=None,
                 cnf={}, fig: Figure | None = None, ax: Axes | None = None,
                 saveas: str = "figure1.png",
                 **kwargs):
        super().__init__(master, cnf, fig=fig, ax=ax, saveas=saveas, **kwargs)

        self.widgets.buttons['linear'] = tk.Button(
            self, text="Linear", command=self.set_fig_to_linear
        )
        self.widgets.buttons['semilogy'] = tk.Button(
            self, text="Semilogy", command=self.set_fig_to_semilogy
        )

        # pack linear and semilogy buttons
        self.widgets.buttons['linear'].pack(
            side=tk.LEFT, fill=tk.BOTH, expand=True
        )
        self.widgets.buttons['semilogy'].pack(
            side=tk.RIGHT, fill=tk.BOTH, expand=True
        )


class LinearSemilogyDoubleCanvas(tk.Frame):
    def __init__(self, master=None, cnf=None, saveas="figures.png", sappends=["_linear", "_semilogy"], n=2, sharex=True, **kwargs):
        width = kwargs.pop("width", 5)
        height = kwargs.pop("height", 4)
        dpi = kwargs.pop("dpi", 100)
        figure_kw = {'width': width, 'height': height, 'dpi': dpi}

        super().__init__(master, cnf, **kwargs)

        if isinstance(saveas, list):
            self.saveass = saveas
        elif isinstance(saveas, str):
            slist = saveas.split('.')
            if len(slist) == 1:
                path = saveas
                ext = str()
            elif len(slist) >= 2:
                num_ext_char = len(slist[-1])
                iend = -(num_ext_char+1)
                ext = saveas[iend:]
                path = saveas[:iend]
            elif len(slist) == 0:
                raise ValueError("Something went wrong")
            else:
                raise ValueError("Lenth of list cannot be negative")

            saveas = []
            for i in range(n):
                if i == 0:
                    sappend = "_linear"
                elif i == 1:
                    sappend = "_semilogy"
                else:
                    sappend = str(i)
                saveas.append(path+sappend+ext)

        self.widgets = Widgets()

        self.widgets.frames['linear'] = MatplotlibCanvasWToolbarSave(
            self, saveas=saveas[0], **figure_kw)

        self.widgets.frames['semilogy'] = MatplotlibCanvasWToolbarSave(

            self, saveas=saveas[1], **figure_kw)
        # remap ax for plotting
        self.frame1 = self.widgets.frames['linear']
        self.ax1 = self.frame1.ax
        self.frame2 = self.widgets.frames['semilogy']
        self.ax2 = self.frame2.ax
        # set frame2 to semilogy
        self.widgets.frames['semilogy'].set_fig_to_semilogy()

        if sharex:
            self.ax2.sharex(self.ax1)

        # pack
        self.frame1.grid(row=0, column=0, sticky="NSWE")
        self.frame2.grid(row=0, column=1, sticky="NSWE")

        # add save default button
        self.widgets.buttons['save_both'] = tk.Button(
            self, text="Save Both", command=self.save_both)
        self.widgets.buttons['save_both'].grid(
            row=1, column=0, columnspan=2, sticky="NSEW")

    def save_both(self, **kwargs):

        self.frame1.save()
        self.frame2.save()

    @staticmethod
    def update_both(func, *args, **kwargs):
        def decorator(self, *args, **kwargs):
            func(*args, **kwargs)
            self.frame1.set_fig_to_linear()
            self.frame2.set_fig_to_semilogy()
        return decorator  # (*args, **kwargs)

    @update_both
    @staticmethod
    def add_plot(ax_func, *args, **kwargs):
        ax_func(*args, **kwargs)

    @update_both
    @staticmethod
    def add_to_both(ax1_func, ax2_func, *args, **kwargs):
        ax1_func(*args, **kwargs)
        ax2_func(*args, **kwargs)

    def legend(self, ax1_bool=True, ax2_bool=True, *args, **kwargs):
        if ax1_bool:
            self.ax1.legend(*args, **kwargs)
        if ax2_bool:
            self.ax2.legend(*args, **kwargs)

    def add_raw_data(self, voltage, current, *args, **kwargs):
        # set defaults
        kwargs.setdefault("label", "Raw Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "s")
        kwargs.setdefault("markerfacecolor", "none")
        kwargs.setdefault("alpha", 0.5)

        # save data
        self.raw_data = pd.DataFrame({'voltage': voltage, 'current': current})

        # add plot
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)

    def add_data(self, voltage, current, *args, **kwargs):
        # set defaults
        kwargs.setdefault("label", "Experimental Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "^")
        kwargs.setdefault("markerfacecolor", "none")

        # save data
        self.data = pd.DataFrame({'voltage': voltage, 'current': current})

        # add plot
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)

    def add_deleted_points(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Removed Data")
        kwargs.setdefault("color", "C1")
        kwargs.setdefault("linestyle", "none")
        kwargs.setdefault("marker", "x")
        self.deleted_points = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)

    def add_filtered_data(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Filtered Data")
        kwargs.setdefault("color", "C7")
        kwargs.setdefault("marker", "v")
        kwargs.setdefault("markerfacecolor", "none")
        self.filtered_data = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)

    def add_smoothed_data(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Smoothed Data")
        kwargs.setdefault("color", "C8")
        kwargs.setdefault("marker", "o")
        kwargs.setdefault("markerfacecolor", "none")
        self.smoothed_data = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_to_both(self.ax1.plot, self.ax2.plot,
                         voltage, current, *args, **kwargs)


class LinearSemilogyDoubleCanvasSingleProbeLangmuir(LinearSemilogyDoubleCanvas):
    def __init__(self, master=None, cnf=None, saveas="figures.png", sappends=["_linear", "_semilogy"], n=2, sharex=True, **kwargs):
        super().__init__(master, cnf, saveas, sappends, n, sharex, **kwargs)

        self.raw_data = None
        self.filtered_data = None
        self.deleted_points = None
        self.smoothed_data = None

        self.vf = None
        self.vs = None

        self.electron_retarding_fits = []
        self.electron_retarding_fills = []

        self.electron_saturation_fit = None
        self.electron_saturation_fill = None

        self.ion_saturation_fit = None
        self.ion_saturation_fill = None

    def set_xlabel_linear(self, label=None, *args, **kwargs):
        label = r"$V_{bias}\ [V]$" if label is None else label
        self.ax1.set_xlabel(label)

    def set_xlabel_semilogy(self, label=None, *args, **kwargs):
        label = r"$V_{bias}\ [V]$" if label is None else label
        self.ax2.set_xlabel(label)

    def set_xlabel(self, label_linear=None, label_semilogy=None, *args, **kwargs):
        self.set_xlabel_linear(label=label_linear)
        self.set_xlabel_semilogy(label=label_semilogy)

    def set_ylabel_linear(self, label=None, *args, **kwargs):
        label = r"$I_{probe}\ [A]$" if label is None else label
        self.ax1.set_ylabel(label)

    def set_ylabel_semilogy(self, label=None, *args, **kwargs):
        label = r"$I_{e}\ [A]$" if label is None else label
        self.ax2.set_ylabel(label)

    def set_ylabel(self, label_linear=None, label_semilogy=None, *args, **kwargs):
        self.set_ylabel_linear(label=label_linear)
        self.set_ylabel_semilogy(label=label_semilogy)

    def add_probe_trace(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Experimental Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "^")
        kwargs.setdefault("markerfacecolor", "none")
        # save data
        self.probe_trace = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_plot(self.ax1.plot,
                      voltage, current, *args, **kwargs)
        self.set_xlabel_linear()
        self.set_ylabel_linear()

    def add_electron_trace(self, voltage, current, *args, **kwargs):
        kwargs.setdefault("label", "Experimental Data")
        kwargs.setdefault("color", "C0")
        kwargs.setdefault("marker", "^")
        kwargs.setdefault("markerfacecolor", "none")
        # save data
        self.electron_trace = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_plot(self.ax2.plot,
                      voltage, current, *args, **kwargs)
        self.set_xlabel_semilogy()
        self.set_ylabel_semilogy()

    def add_floating_potential(self, V_f, *args, **kwargs):
        kwargs.setdefault("label", rf"$V_{{f}} = {V_f:.2f}V$")
        kwargs.setdefault("color", "C2")
        self.V_f = V_f
        self.add_to_both(self.ax1.axvline, self.ax2.axvline,
                         V_f, *args, **kwargs)

    def add_plasma_potential(self, V_s, *args, **kwargs):
        kwargs.setdefault("label", rf"$V_{{s}} = {V_s:.2f}V$")
        kwargs.setdefault("color", "C4")
        self.V_s = V_s
        self.add_to_both(self.ax1.axvline, self.ax2.axvline,
                         V_s, *args, **kwargs)

    def add_electron_retarding_fit(self, voltage, current, KT_e, *args, **kwargs):
        kwargs.setdefault("label", rf"$KT_{{e}} = {KT_e:.2f}eV$")
        kwargs.setdefault("color", "C3")
        kwargs.setdefault("linestyle", "--")
        self.electron_retarding_fits.append(
            pd.DataFrame({'voltage': voltage, 'current': current}))
        self.add_plot(self.ax2.plot,
                      voltage, current, *args, **kwargs)

    def add_ion_saturation_fit(self, voltage, current, I_is, *args, **kwargs):
        kwargs.setdefault("label", rf"$I_{{is}} = {I_is:.2e}A$")
        kwargs.setdefault("color", "C6")
        kwargs.setdefault("linestyle", "--")
        self.ion_saturation_fit = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_plot(self.ax1.plot,
                      voltage, current, *args, **kwargs)

    def add_electron_saturation_fit(self, voltage, current, I_es, *args, **kwargs):
        kwargs.setdefault("label", rf"$I_{{es}} = {I_es:.2e}A$")
        kwargs.setdefault("color", "C5")
        kwargs.setdefault("linestyle", "--")
        self.electron_saturation_fit = pd.DataFrame(
            {'voltage': voltage, 'current': current})
        self.add_plot(self.ax2.plot,
                      voltage, current, *args, **kwargs)

    def add_electron_retarding_fill(self, xstart, xstop, *args, **kwargs):
        kwargs.setdefault("color", "C3")
        kwargs.setdefault("alpha", 0.3)
        self.electron_retarding_fills.append([xstart, xstop])
        self.add_to_both(self.ax1.axvspan, self.ax2.axvspan,
                         xstart, xstop, *args, **kwargs)

    def add_electron_saturation_fill(self, xstart, xstop, *args, **kwargs):
        kwargs.setdefault("color", "C5")
        kwargs.setdefault("alpha", 0.3)
        self.electron_saturation_fill = [xstart, xstop]
        self.add_to_both(self.ax1.axvspan, self.ax2.axvspan,
                         xstart, xstop, *args, **kwargs)

    def add_ion_saturation_fill(self, xstart, xstop, *args, **kwargs):
        kwargs.setdefault("color", "C6")
        kwargs.setdefault("alpha", 0.3)
        self.ion_saturation_fill = [xstart, xstop]
        self.add_to_both(self.ax1.axvspan, self.ax2.axvspan,
                         xstart, xstop, *args, **kwargs)

    def make_plot(self, solver):
        app = self
        # make xdomain for fit plotting
        xdomain = [np.min(solver.data.voltage), np.max(solver.data.voltage)]

        # solve for KT_e fit data
        KT_e_fit = solver.results["KT_e"]["other"]["fit"]
        voltage_KT_e_fit = KT_e_fit.xrange(domain=xdomain)
        current_KT_e_fit = KT_e_fit(voltage_KT_e_fit)
        # solve for I_es fit data
        I_es_fit = solver.results["I_es"]["other"]["fit"]
        voltage_I_es_fit = I_es_fit.xrange(domain=xdomain)
        current_I_es_fit = I_es_fit(voltage_I_es_fit)

        # solve for I_is fit data
        I_is_fit = solver.results["I_is"]["other"]["fit"]
        voltage_I_is_fit = I_is_fit.xrange(domain=xdomain)
        current_I_is_fit = I_is_fit(voltage_I_is_fit)

        # Make Plots ###
        # add probe trace
        app.add_probe_trace(solver.data.voltage, solver.data.current)
        # add electron current
        app.add_electron_trace(solver.data.voltage, solver.data.current_e)
        # mark deleted points
        app.add_deleted_points(solver.deleted_data.voltage,
                               solver.deleted_data.current)

        # Add to Plasma Properties to Plots
        app.add_floating_potential(solver.results["V_f"]["value"])
        app.add_plasma_potential(solver.results["V_s"]["value"])

        # Electron Retarding fit and fill for region of fit
        app.add_electron_retarding_fit(
            voltage_KT_e_fit, current_KT_e_fit, solver.results["KT_e"]["value"])
        app.add_electron_retarding_fill(*KT_e_fit.poly.domain)

        # Electron Saturation fit and fill for region of fit
        app.add_electron_saturation_fit(
            voltage_I_es_fit, current_I_es_fit, solver.results["I_es"]["value"])
        app.add_electron_saturation_fill(*I_es_fit.poly.domain)

        # Ion Saturation fit and fill for region of fit or domain area
        app.add_ion_saturation_fit(
            voltage_I_is_fit, current_I_is_fit, solver.results["I_is"]["value"],
        )
        app.add_ion_saturation_fill(*I_is_fit.poly.domain)

        # turn on legend
        app.legend()


def other():
    # Make save all button
    # btn_save_all = tk.Button(master=self, text="Save Both and Next",
    #                         command=lambda: save_figures(fig, saveas, root))
    # btn_save_all.pack(side=tk.BOTTOM, expand=1, fill=tk.BOTH)

    if None:
        canvas.mpl_connect(
            "key_press_event", lambda event: print(f"you pressed {event.key}"))
        canvas.mpl_connect("key_press_event", key_press_handler)

        # Make quit button
        button_quit = tk.Button(master=root, text="Quit", command=root.destroy)
        button_quit.pack(side=tk.BOTTOM, expand=1, fill=tk.BOTH)


class SingleProbeLangmuirResultsFrame(tk.Frame):
    def __init__(self, master=None, cnf={}, *args, results=None, **kwargs):
        super().__init__(master, cnf, *args, **kwargs)

        # add widgets organizer
        self.widgets = Widgets()

        # if results are not given
        if results is None:
            results = {}
            for key in available_plasma_properties:
                results[key] = {"value": 1, "other": None}
        elif isinstance(results, dict):
            pass
        else:
            raise ValueError(
                "'results' must be dictionary: {}".format(type(results)))

        # creat instance
        self._results = results

        # number of properties to display
        grid_positions_list = self.grid_positions(results.keys())

        # loop and create labels to display
        labels = {}
        for k, key in enumerate(results.keys()):
            text = self.results[key]["value"]
            unit = self.results[key]["other"]
            unit = self.default_unit_format(key) if unit is None else unit
            labels[key] = ResultOutput(key, text, unit, self, cnf)
            row = grid_positions_list[k][0]
            col = grid_positions_list[k][1]
            labels[key].grid(row=row, column=col,
                             sticky="NWSE", padx=5, pady=5)

        self.widgets.frames.update(labels)

        self.update_texts(results=self.results)

        self.columnconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=1)

    @property
    def results(self):
        return self._results

    @results.setter
    def results(self, val):
        self.update_texts(val)
        self._results = val

    def update_label(self, key, text):
        if key in self.widgets.frames:
            self.widgets.frames[key].update_text(text=text)
        else:
            table = "\n".join(
                [f"{k}\t{v}" for k, v in enumerate(self.widgets.frames)])
            raise ValueError(
                f"Matching key not found: {key}\nChoose from one of the available options:\n{table}")

    def update_texts(self, results=None):
        if results is None:
            results = self.results

        for key in results.keys():
            text = self.default_text_format(key)
            val = results[key]['value']
            val = 0.0 if val is None else val
            text = text.format(val)
            self.update_label(key, text)

    def default_text_format(self, key):
        if key == "V_f":
            text = "{0:.2f}"
        elif key == "V_s":
            text = "{0:.2f}"
        elif key == "I_is":
            text = "{0:.2e}"
        elif key == "n_i":
            text = "{0:.2e}"
        elif key == "I_es":
            text = "{0:.2e}"
        elif key == "KT_e":
            text = "{0:.2f}"
        elif key == "n_e":
            text = "{0:.2e}"
        elif key == "J_es":
            text = "{0:.2e}"
        elif key == "J_is":
            text = "{0:.2e}"
        elif key == "lambda_De":
            text = "{0:.2e}"
        elif key == "sheath" or key == "r_p/lambda_De":
            text = "{0:.2e}"
        else:
            text = "{}"

        return text

    def default_unit_format(self, key):
        if key == "V_f":
            text = "V"
        elif key == "V_s":
            text = "V"
        elif key == "I_is":
            text = "A"
        elif key == "n_i":
            text = "m\u207B\u00B3"
        elif key == "I_es":
            text = "A"
        elif key == "KT_e":
            text = "eV"
        elif key == "n_e":
            text = "m\u207B\u00B3"
        elif key == "J_es":
            text = "A/m\u00B2"
        elif key == "J_is":
            text = "A/m\u00B2"
        elif key == "lambda_De":
            text = "m"
        elif key == "sheath" or key == "r_p/lambda_De":
            text = ""
        else:
            text = "N/A"

        return text

    def grid_positions(self, lst, layout="two"):
        # layout = "square"
        if layout == "two":
            func = self._two_layout
        elif layout == "square":
            func = self._square_layout
        else:
            raise ValueError("layout is not knonw: {layout}")
        return func(lst)

    def _sort_positions(self, num_items, num_rows, num_cols):
        positions = []
        for i in range(num_items):
            row = int(i/num_cols)
            col = i % num_cols
            positions.append((row, col))

        return positions

    def _two_layout(self, lst):
        num_items = len(lst)
        num_rows = int(np.ceil(num_items/2))
        num_cols = 2
        return self._sort_positions(num_items, num_rows, num_cols)

    def _square_layout(self, lst):
        # Calculate the number of rows and columns needed for the closest square grid
        num_items = len(lst)
        grid_size = int(np.ceil(np.sqrt(num_items)))
        num_rows = grid_size
        num_cols = grid_size

        # If the grid is larger than necessary, reduce the number of columns
        if num_items < num_rows*num_cols:
            num_cols = int(np.ceil(num_items/num_rows))

        return self._sort_positions(num_items, num_rows, num_cols)


class ResultOutput(tk.Frame):
    def __init__(self, key, value=None, unit=None, master=None, cnf=None, *args, latex=True, **kwargs):
        super().__init__(master, cnf, *args, **kwargs)

        # create widgets
        self.widgets = Widgets()

        # create button/label of property name to display
        btn_kwargs = kwargs.get("btn_kwargs", {})
        if latex is True:
            key = self.convert_to_latex(key)
        btn_kwargs.setdefault("text", key)
        # define later to call new window with settings
        btn_kwargs.setdefault("command", None)
        btn = tk.Button(self, cnf, **btn_kwargs)

        # create label of value
        value_lbl_kwargs = kwargs.get("value_lbl_kwargs", {})
        value_lbl_kwargs.setdefault("text", value)
        value_lbl_kwargs.setdefault("bg", "white")
        value_lbl = tk.Label(self, cnf, **value_lbl_kwargs)

        # create unit label
        unit_lbl_kwargs = kwargs.get("unit_lbl_kwargs", {})
        unit_lbl_kwargs.setdefault("text", unit)
        unit_lbl = tk.Label(self, cnf, **unit_lbl_kwargs)

        # pack away!
        btn.grid(row=0, column=0, sticky="NSWE")
        value_lbl.grid(row=0, column=1, sticky="NSWE")
        unit_lbl.grid(row=0, column=2, sticky="NSWE")

        # save to self
        self.widgets.buttons["key"] = btn
        self.widgets.labels["value"] = value_lbl
        self.widgets.labels["unit"] = unit_lbl

        # configure columns same size
        self.columnconfigure(0, weight=2, uniform="fred")
        self.columnconfigure(1, weight=2, uniform="fred")
        self.columnconfigure(2, weight=1, uniform="fred")

    def update_text(self, text):
        self.widgets.labels["value"].config(text=text)

    def updata_value_label(self, key, value):
        kwargs = {key: value}
        self.widgets.labels["value"].config(**kwargs)

    def convert_to_latex(self, input_string: str):

        input_string = input_string.replace("lambda", "\u03BB")
        input_string = input_string.replace("sheath", "r_p/\u03BB_De")

        split_string = input_string.split("_")
        if len(split_string) > 1:
            text = []
            for k, string in enumerate(split_string):
                if k == 0:
                    text.append(string)
                else:
                    text.append(string)
            text = "".join(text)
        else:
            text = split_string[0]

        return text


class CombinedDataFrame(tk.Frame):
    def __init__(self, Data, Solver, Canvas, Panel, master=None, cnf=None, *args, **kwargs):
        super().__init__(master, cnf=cnf, *args, **kwargs)
        # get kwargs for individual
        canvas_kwargs = kwargs.pop("canvas_kwargs", {})
        panel_kwargs = kwargs.pop("panel_kwargs", {})
        solver_kwargs = kwargs.pop("solver_kwargs", {})

        # set defaults for canvas
        canvas_kwargs.setdefault("width", 5)
        canvas_kwargs.setdefault("height", 4)

        # set defaults for panel

        # set defaults for solver
        solver_kwargs.setdefault("Data", Data)

        # make plotting canvas
        self._canvas = Canvas(  # LinearSemilogyDoubleCanvasSingleProbeLangmuir(
            self, cnf, **canvas_kwargs)

        # make control panel
        self._panel = Panel(  # SingleProbeLangmuirResultsFrame(
            self, cnf, **panel_kwargs)

        # make solver
        self._solver = Solver(**solver_kwargs)

        # pack Away!!
        self.canvas.grid(row=0, column=1, sticky="NSWE")
        self.panel.grid(row=0, column=0, sticky="NWE")

    @property
    def canvas(self):
        return self._canvas

    @property
    def panel(self):
        return self._panel

    @property
    def solver(self):
        return self._solver


class LinearSemilogySingleLangmuirCombinedData(CombinedDataFrame):
    def __init__(self, Data: pd.DataFrame, master=None, cnf=None, *args, **kwargs):
        Canvas = LinearSemilogyDoubleCanvasSingleProbeLangmuir
        Panel = SingleProbeLangmuirResultsFrame
        Solver = SingleLangmuirProbeSolver
        super().__init__(Data, Solver, Canvas, Panel, master, cnf=cnf, *args, **kwargs)


class MultipleDataFrame(tk.Frame):
    def __init__(self, num, DataFrame, master=None, cnf=None, *args, **kwargs):
        super().__init__(master, cnf=cnf, *args, **kwargs)

        self._pages = [None]*num

    def create_pages(self, num, DataFrame, *args, **kwargs):
        for k in range(num):
            self.pages[k] = DataFrame(*args, **kwargs)

    @property
    def pages(self):
        return self._pages

    def show_page(self, n):
        n -= 1

        self.pages[n].tkraise()



from pstl.utls.protocol.gpib import initialize as init

class Open():
    def __init__(self,port=None):

        # trys to open if given port
        # if fails, it gives you options
        if port is not None:
            try:
                res=init.open_port(port)
            except:
                print("\nFailed to open %s"%(port))
                port=None
        if port is None:
            res=init.choose_port()


        self.visa=res

        self.port=self.visa.resource_name
        
        self.write=self.visa.write
        self.read=self.visa.read
        self.query=self.visa.query



import tkinter as tk
import argparse

import numpy as np
import pandas as pd
from matplotlib import style

from pstl.gui.langmuir import LinearSemilogyCanvas, LinearSemilogyDoubleCanvas, LinearSemilogyDoubleCanvasSingleProbeLangmuir
from pstl.gui.langmuir import LinearSemilogyDoubleCanvasSingleProbeLangmuir as Canvas
from pstl.gui.langmuir import SingleProbeLangmuirResultsFrame as Panel
from pstl.gui.langmuir import SingleProbeLangmuirResultsFrame
from pstl.gui.langmuir import LinearSemilogySingleLangmuirCombinedData as LSSLCD

from pstl.utls.plasmas import XenonPlasma, ArgonPlasma
from pstl.diagnostics.probes.langmuir.single import SphericalSingleProbeLangmuir as SSPL
from pstl.diagnostics.probes.langmuir.single import CylindericalSingleProbeLangmuir as CSPL
from pstl.diagnostics.probes.langmuir.single.analysis.solver import SingleLangmuirProbeSolver as SLPS

style.use("bmh")

parser = argparse.ArgumentParser(
                    prog='GUI Langmuir',
                    description='Anaylsis on IV-trace',
                    epilog='Text at the bottom of help')
parser.add_argument('-s','--sname', help="save name for plots", default="outpng.png")
parser.add_argument('-f','--fname', help="file name to read in", default="lang_data.txt")
args = parser.parse_args()

def old_main():
    x = np.linspace(-.2, .2, 9)
    y = x*.2

    root = tk.Tk()
    a = LinearSemilogyDoubleCanvasSingleProbeLangmuir(root, width=6, height=5)
    results = None
    b = SingleProbeLangmuirResultsFrame(root, results=results)
    a.add_raw_data(x, y)
    a.add_floating_potential(.2, color="r")
    a.add_electron_retarding_fill(-0.1, .1, color="black", alpha=0.1)
    a.legend()
    # a = LinearSemilogyCanvas(root)
    # a.widgets.frames['semilogy'].ax.relim()
    # a.widgets.frames['semilogy'].ax.autoscale()
    # a.widgets.frames['semilogy'].set_fig_to_semilogy()
    # a.ax2.update_from(a.ax1)
    # a.widgets.frames['semilogy'].fig.canvas.draw()
    # a.widgets.frames['linear'].fig.canvas.draw()

    # pack
    a.grid(row=0, column=1)
    b.grid(row=0, column=0)

    # run loop
    root.mainloop()


def get_lang_data():
    filename = args.fname
    data = pd.read_csv(filename, names=["voltage", "current"])
    data.iloc[:, 1] *= 1
    return data


def main():
    # initiate app
    app = tk.Tk()

    # create page
    page = tk.Frame(app)
    page.pack()

    # create Canvas
    canvas = Canvas(page, saveas=args.sname, width=5, height=4)
    canvas.grid(row=0, column=1, sticky="NSWE", rowspan=2)

    # create Panel
    panel = Panel(page)
    panel.grid(row=0, column=0, sticky="N")

    data = get_lang_data()
    # data = get_test_data()

    # probe = CSPL(2.4e-3, 10e-3)
    probe = SSPL(0.010, 0.0079)
    probe_smaller = SSPL(0.010, 0.009)
    rox_probe = CSPL(0.76e-3,2.54e-3)
    plasma = XenonPlasma()
    # plasma = ArgonPlasma()

    solver = SLPS(plasma, rox_probe, data)

    # pre process data
    solver.preprocess()

    # solve for data
    solver.find_plasma_properties()

    # make plots of solved for plasma parameters
    canvas.make_plot(solver)

    # pass data to results panel
    panel.results = solver.results

    # run mainloop
    app.mainloop()


if __name__ == "__main__":
    main()




import numpy as np
from numpy.polynomial import Polynomial as P

from pstl.utls import constants as c
from pstl.utls.plasmas.sheath import thick, transitional
from pstl.utls.helpers import method_selector, method_function_combiner


def get_ion_density(*args, method=0, **kwargs):
    # set default -> (thin)
    method = 0 if method is None else method

    # select function based on integer_method
    func = method_selector(method, available_methods, available_functions)

    # Call funtion and return result
    return func(*args, **kwargs)


def thin_sheath_method(I_is, area, KT_e, m_i, shape="Unknown", *args, **kwargs):
    # KT_e [eV]
    inv_root = np.sqrt(m_i/(KT_e*c.e))
    # np.exp(-0.5) = 1/np.exp(0.5) = 0.61
    n_i = np.abs(I_is)*inv_root/(area*c.q_e)*np.exp(0.5)
    return n_i, {"method": "thin", "shape": shape}


def transitional_sheath_method(voltage, current, area, m_i, KT_e, shape, r_p, lambda_D, *args, **kwargs):
    func = transitional.shape_selector(shape)
    a, b = func(r_p, lambda_D)
    n_i = _transitional_and_thick_method(
        voltage, current, area, m_i, KT_e, a, b)
    return n_i, {"method": "transitional", "shape": shape}


def thick_sheath_method(voltage, current, area, m_i, KT_e, shape, *args, **kwargs):
    func = thick.shape_selector(shape)
    a, b = func()
    n_i = _transitional_and_thick_method(
        voltage, current, area, m_i, KT_e, a, b)
    return n_i, {"method": "thick", "shape": shape}


def _transitional_and_thick_method(voltage, current, area, m_i, KT_e, a, b):
    A = 1/(a*area)*np.sqrt(2*np.pi*m_i) * \
        np.power(c.e, -1.5)*np.power(KT_e, b-0.5)
    mod_I_probe = np.power(current, 1/b)
    polyfit = P.fit(voltage, mod_I_probe, deg=1)
    slope = polyfit.convert().coef[-1]
    n_i = np.multiply(A, np.power(slope, b))
    return n_i


# Declare available methods
available_methods = {
    0: ['thin'],
    1: ['transitional'],
    2: ['thick', 'OML'],
}
# Declare correspondin functions for available_methods
available_functions = {
    0: thin_sheath_method,
    1: transitional_sheath_method,
    2: thick_sheath_method,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
options = method_function_combiner(available_methods, available_functions)




from pstl.instruments.daq.agilent import utls as sul

class Agilent34901A():
    def __init__(self,slot=None,nchannels=20,chtype=None):

        """
        creates a object for the card agilent 34901A.
        20 Channels 
        (Note channel[0] is not an acual channel but will 
        display the list of channels and thier configured type.)
        """
        self.name="agilent34901A"
        self.type="agilent card"
        channels = [None]*(nchannels+1)
        # check of chtypes was a list
        channels = sul.loop_setup(channels,chtype,slot)
            
        self.nchannels=nchannels
        self.slot=slot
        channels[0] = self.list_channels
        self.channel = channels

    def list_channels(self):
        nchannels=self.nchannels
        print()
        for k in range(1,nchannels+1):
            channel=self.channel[k].location
            chtype=self.channel[k].chtype
            print("Channel %s: Type: %s\n"%(str(channel),str(chtype)))








from .ion_density import get_ion_density
from .electron_density import get_electron_density
from .plasma_potential import get_plasma_potential
from .electron_saturation_current import get_electron_saturation_current, get_electron_saturation_current_density
from .electron_temperaure import get_electron_temperature
from .ion_saturation_current import get_ion_saturation_current, get_ion_saturation_current_density
from .floating_potential import get_floating_potential
from .ion_current import get_ion_current
from pstl.utls.plasmas.sheath import get_probe_to_sheath_ratio
from typing import Dict, Any

import numpy as np
import pandas as pd

from pstl.utls import constants as c
from pstl.utls.constants import lambda_D as get_lambda_D
from pstl.utls.decorators import absorb_extra_args_kwargs, add_empty_dict
get_lambda_D = add_empty_dict(absorb_extra_args_kwargs(get_lambda_D))

default_plasma_properties_to_get = [
    "V_f",
    "V_s",
    "KT_e",
    "n_e",
    "I_es",
    "J_es",
    "n_i",
    "I_is",
    "J_is",
    "lambda_De",
    # "r_p/lambda_De",
    "sheath",
]
default_methods = {
    "V_f": "consecutive",
    "V_s": "intersection",
    "I_is": 0,
    "n_i": 0,
    "I_es": 0,
    "KT_e": 0,
    "n_e": 0,
    "J_es": 0,
    "J_is": 0,
    "lambda_De": 0,
    # "r_p/lambda_De": 0,
    "sheath": 0,
}


def print_results(results):
    for key, value in results.items():
        print(f"{key}:")
        for inner_key, inner_value in value.items():
            if key in ["n_i", "n_e", "I_is", "I_es", "J_es", "J_is", "lambda_De"] and inner_key == "value":
                if inner_value is not None:
                    print(f"\t{inner_key}: {inner_value:.2e}")
                else:
                    print(f"\t{inner_key}: {inner_value}")
            elif key in ["V_f", "V_s", "KT_e"] and inner_key == "value":
                if inner_value is not None:
                    print(f"\t{inner_key}: {inner_value:.2f}")
                else:
                    print(f"\t{inner_key}: {inner_value}")
            else:
                print(f"\t{inner_key}: {inner_value}")


def _return_orgainizer(returns):
    # check if tuple for indexing
    if isinstance(returns, tuple):
        value = returns[0]
        other = returns[1:]
    else:
        value = returns
        other = None
    return value, other


def get_plasma_property_sorter(key, method, *args, **kwargs):

    # choose which property
    if key == "V_f":
        func = get_floating_potential
    elif key == "V_s":
        func = get_plasma_potential
    elif key == "I_is":
        func = get_ion_saturation_current
    elif key == "n_i":
        func = get_ion_density
    elif key == "I_es":
        func = get_electron_saturation_current
    elif key == "KT_e":
        func = get_electron_temperature
    elif key == "n_e":
        func = get_electron_density
    elif key == "J_es":
        func = get_electron_saturation_current_density
    elif key == "J_is":
        func = get_ion_saturation_current_density
    elif key == "lambda_De":
        func = get_lambda_D
    elif key == "sheath":  # key == "r_p/lamda_De":
        func = get_probe_to_sheath_ratio
    else:
        table = "\n".join(
            [f"{k}\t{v}" for k, v in enumerate(default_methods)])
        raise ValueError(
            f"Matching key not found: {key}\nChoose from one of the available options:\n{table}")

    # solve and return a tuple
    return func(*args, method=method, **kwargs)


def lobbia():
    return


def _topham(voltage, current, area, m_i, m_e=c.m_e,
            methods={},
            smooth=False, itmax=1, convergence_percent=1,
            *args, **kwargs) -> Dict:
    properties = None
    if not isinstance(methods, dict):
        raise ValueError(
            "'methods' must be a dictionary not: ", type(methods))
    # convert convergence percent to a decimal
    convergence_decimal = convergence_percent/100

    # overwrite methods if passed in
    methods_to_use = dict(default_methods)
    methods_to_use.update(methods)

    # overwrite properties if passed in (not implemented yet)
    if properties is None:
        properties_to_get = list(default_plasma_properties_to_get)
    elif isinstance(properties, list):
        properties_to_get = list(properties)
    else:
        raise ValueError(
            "'properteies' must be a list or None not: ", type(methods))

    # set up results dictionary for returns
    results = {}
    for plasma_property in properties_to_get:
        results[plasma_property] = {'value': None, 'other': None}

    # see what kwargs are given and creat a new dictionary
    func_kwargs = {}
    for key in methods_to_use.keys():
        func_kwargs[key] = kwargs.get(key+"_kwargs", {})

    # Zero Step (optional):
    # smooth data for find better fit region

    # First Step:
    # Get Floating Potential
    # -> V_f
    key = "V_f"
    results[key]["value"], results[key]["other"] = get_plasma_property_sorter(
        key, methods_to_use[key], *args, **func_kwargs[key])

    # Intermediate Step (optional for convergence speedup):
    # Do Either @ V_bias << V_f:
    # 1) Fit a linear line in Ion Saturation Region
    # 2) Take the resonable minimum value of the Ion Saturation Region
    # 3) Fit a power term (i.e. 1/2) to the Ion Saturation Region
    # Then either use the fits to subtract out ion current from the total probe current
    # to get a approximate electron only current, or a flat value across the board of the
    # minimum ion current (may lead to artifcial bi-Maxwellian)
    # Ie = Iprobe - Ii  --or-- Iprobe = Ie+Ii (note: Ii convention is negative)
    # -> temp I_i
    I_i, I_i_extras = get_ion_current(
        shape,
    )

    convergence = False
    it = 0
    vs_old = float("inf")
    key_convergence = "V_s"
    while not convergence and it < itmax:
        # number of iterations
        it += 1

        # Second Step:
        # Find Rough Exponential Fit after Floating Potential in the Electron Retarding Region
        # Note: Ii is removed this should be only Ie
        # -> KT_e

        # Third Step:
        # Find Electron Saturation Exponential Fit
        # @ V_bias >> V_s (plasma or space potential) before probe breakdown (plasma created due to accelerated elcectrons),
        # There should be no ion current.
        # Note: Theory says I_es =
        # -> I_es = I_e(@V_s) = exp(m*V_s+b)

        # Fourth Step:
        # Find Plasma Potential via intersection point of Electron Retarding and Saturation Regions
        # May also be done using first or second derivate to locate like in lobbia but requries smoothed data
        # and often inconclusive
        # -> V_s

        # Fifth Step:
        # Find Ion Saturation Current
        # Either:
        # 1) Thin Sheath:
        #   Linear Ion Saturation Fit @ V_bias << V_f that intersects the x-axis at a V_bias > V_s
        #   -> I_is = I_i(@V_s) = m*V_s + b
        # 2) Thick Sheath/Orbital Motion Limited (OML):
        #   Power 1/2 Fit in the Ion Saturation Region @ V_bias << V_f
        #   I_i^2 = alpha*V_bias + beta
        #       Where: alpha = -(q_e*area*ni)^2*(2.1*q_e/(pi^2*m_i))
        #               beta = (q_e*area*ni)^2*(2.1*q_e/(pi^2*m_i))*V_s
        #   Note: in theory I_es = -I_is*exp(0.5)*sqrt(m_i/(2*pi*m_e))
        # questionable below:
        #   -> I_is = -I_es*exp(-0.5)*sqrt(2*pi*m_e/m_i) --or-- I_is = -exp(-0.5)*area*q_e*n_i*sqrt(KT_e/m_i)

        # combines while loop steps into a few lines
        for key in ["KT_e", "I_es", "V_s", "I_is"]:
            results[key]["value"], results[key]["other"] = get_plasma_property_sorter(
                key, methods_to_use[key], *args, **func_kwargs[key])

        # Repeat till Convergence on V_s from Intermediate Step to Last Step
        # using Ion Saturation Current fit to correct orginal probe current data to electron only current
        # i.e. I_probe,orginal - I_i,new = I_e,next_iteration_data

        vs_new = results[key_convergence]['value']
        difference = vs_old-vs_new
        realtive_difference = difference/vs_old
        abs_rel_diff = np.abs(realtive_difference)

        convergence = True if abs_rel_diff <= convergence_decimal else False

    # Last Step:
    # Once convergence is made, get n_i and n_e (ion and electron densities) and J_es and J_is (electorn and ion saturation current densities)
    # Electrons:
    # -> n_e = I_es/(area*q_e)*sqrt(2*pi*m_e/KT_e)
    # -> J_es = I_es/area
    # Ions:
    # if thin sheath:
    # n_i = I_is/(area*q_e)*sqrt(m_i/KT_e)
    # if thick sheath:
    # n_i = ((alpha*pi^2*m_i)/(2*q_e^3*area^2))^(0.5)

    # Debye Length
    # -> lambda_De = sqrt(KT_e*epsilon_0/(n_e*q_e^2))

    # combines while loop steps into a few lines
    for key in ["n_e", "n_i", "J_es", "J_is", "lambda_De"]:
        results[key]["value"], results[key]["other"] = get_plasma_property_sorter(
            key, methods_to_use[key], *args, **func_kwargs[key])

    return results

######################################################################


def topham(voltage, current, shape, r_p, area, m_i, m_e=c.m_e,
           methods={},
           smooth=False, itmax=1, convergence_percent=1,
           *args, **kwargs) -> Dict:
    properties = None
    if not isinstance(methods, dict):
        raise ValueError(
            "'methods' must be a dictionary not: ", type(methods))
    # convert convergence percent to a decimal
    convergence_decimal = convergence_percent/100

    # overwrite methods if passed in
    methods_to_use = dict(default_methods)
    methods_to_use.update(methods)

    # overwrite properties if passed in (not implemented yet)
    if properties is None:
        properties_to_get = list(default_plasma_properties_to_get)
    elif isinstance(properties, list):
        properties_to_get = list(properties)
    else:
        raise ValueError(
            "'properteies' must be a list or None not: ", type(methods))

    # set up results dictionary for returns
    results = {}
    for plasma_property in properties_to_get:
        results[plasma_property] = {'value': None, 'other': None}

    # see what kwargs are given and creat a new dictionary
    func_kwargs = {}
    for key in methods_to_use.keys():
        func_kwargs[key] = kwargs.get(key+"_kwargs", {})

    # inialize sheath method
    sheath_method = "thin"

    # Zero Step (optional):
    # smooth data for find better fit region

    # First Step:
    # Get Floating Potential
    # -> V_f
    key = "V_f"
    V_f, V_f_extras = get_floating_potential(
        voltage, current, method="consecutive", interpolate="linear",
    )

    # Intermediate Step (optional for convergence speedup):
    # Do Either @ V_bias << V_f:
    # 1) Fit a linear line in Ion Saturation Region
    # 2) Take the resonable minimum value of the Ion Saturation Region
    # 3) Fit a power term (i.e. 1/2) to the Ion Saturation Region
    # Then either use the fits to subtract out ion current from the total probe current
    # to get a approximate electron only current, or a flat value across the board of the
    # minimum ion current (may lead to artifcial bi-Maxwellian)
    # Ie = Iprobe - Ii  --or-- Iprobe = Ie+Ii (note: Ii convention is negative)
    # -> temp I_i
    I_i, I_i_extras = get_ion_current(
        shape,
        voltage, current,
        method=sheath_method, V_f=V_f,
    )
    I_i_fit = I_i_extras["fit"]

    I_e = np.subtract(current, I_i)

    convergence = False
    it = 0
    old = float("inf")
    while not convergence and it < itmax:
        # number of iterations
        it += 1

        # Second Step:
        # Find Rough Exponential Fit after Floating Potential in the Electron Retarding Region
        # Note: Ii is removed this should be only Ie
        # -> KT_e
        KT_e, KT_e_extras = get_electron_temperature(
            voltage, I_e, method="fit", V_f=V_f,
        )
        KT_e_poly = KT_e_extras["fit"].poly.convert()

        # Third Step:
        # Find Electron Saturation Exponential Fit
        # @ V_bias >> V_s (plasma or space potential) before probe breakdown (plasma created due to accelerated elcectrons),
        # There should be no ion current.
        # Note: Theory says I_es =
        # -> I_es = I_e(@V_s) = exp(m*V_s+b)
        I_es, I_es_extras = get_electron_saturation_current(
            voltage, I_e, method="fit", elec_ret_poly=KT_e_poly, V_f=V_f,
        )
        I_es_poly = I_es_extras["fit"].poly.convert()

        # Fourth Step:
        # Find Plasma Potential via intersection point of Electron Retarding and Saturation Regions
        # May also be done using first or second derivate to locate like in lobbia but requries smoothed data
        # and often inconclusive
        # -> V_s
        V_s, V_s_extras = get_plasma_potential(
            voltage, I_e, method="intersection", V_f=V_f, elec_ret_poly=KT_e_poly, elec_sat_poly=I_es_poly,
        )

        # Fifth Step:
        # Find Ion Saturation Current
        # Either:
        # 1) Thin Sheath:
        #   Linear Ion Saturation Fit @ V_bias << V_f that intersects the x-axis at a V_bias > V_s
        #   -> I_is = I_i(@V_s) = m*V_s + b
        # 2) Thick Sheath/Orbital Motion Limited (OML):
        #   Power 1/2 Fit in the Ion Saturation Region @ V_bias << V_f
        #   I_i^2 = alpha*V_bias + beta
        #       Where: alpha = -(q_e*area*ni)^2*(2.1*q_e/(pi^2*m_i))
        #               beta = (q_e*area*ni)^2*(2.1*q_e/(pi^2*m_i))*V_s
        #   Note: in theory I_es = -I_is*exp(0.5)*sqrt(m_i/(2*pi*m_e))
        # questionable below:
        #   -> I_is = -I_es*exp(-0.5)*sqrt(2*pi*m_e/m_i) --or-- I_is = -exp(-0.5)*area*q_e*n_i*sqrt(KT_e/m_i)
    # Last Step:
    # Once convergence is made, get n_i and n_e (ion and electron densities) and J_es and J_is (electorn and ion saturation current densities)
    # Electrons:
    # -> n_e = I_es/(area*q_e)*sqrt(2*pi*m_e/KT_e)
    # -> J_es = I_es/area
    # Ions:
    # if thin sheath:
    # n_i = I_is/(area*q_e)*sqrt(m_i/KT_e)
    # if thick sheath:
    # n_i = ((alpha*pi^2*m_i)/(2*q_e^3*area^2))^(0.5)

    # Debye Length
    # -> lambda_De = sqrt(KT_e*epsilon_0/(n_e*q_e^2))

        n_e, n_e_extras = get_electron_density(
            I_es, area, KT_e, m_e=c.m_e, method="I_es",
        )
        # get ionsaturation value at flaoting
        I_is, I_is_extras = get_ion_saturation_current(
            voltage, current, V_f=V_f, method="floating",
            I_i_fit=I_i_fit, I_i_method=sheath_method,
        )
        # temp ion density
        n_i, n_i_extras = get_ion_density(
            I_is, area, KT_e, m_i, method=sheath_method,
        )
        # lambda_D
        lambda_De, lambda_De_extras = get_lambda_D(n_e, KT_e)

        # ratio
        ratio, ratio_extras = get_probe_to_sheath_ratio(r_p, lambda_De)
        # sheath_method = ratio_extras["sheath"]

        # Repeat till Convergence on V_s from Intermediate Step to Last Step
        # using Ion Saturation Current fit to correct orginal probe current data to electron only current
        # i.e. I_probe,orginal - I_i,new = I_e,next_iteration_data
        if ratio <= 3:
            sheath_method = "thick"
        elif ratio > 3 and ratio < 50:
            sheath_method = "transitional"
            # get
        elif ratio >= 50 and it == 1:
            sheath_method = "thin"
            break
        elif ratio >= 50:
            errmsg = "Back to thin"
            raise ValueError(errmsg)
        else:
            errmsg = "WHa???!!!"
            raise ValueError(errmsg)

        if it == itmax:
            break

        # get updated n_i
        # <<enter here>>
        print(f"n_i before: {n_i:.2e}")
        n_i, n_i_extras = get_ion_density(
            voltage, current, area, m_i, KT_e, shape, r_p, lambda_De, shape, method=sheath_method)
        print(f"n_i after: {n_i:.2e}")

        # get new ion current with transitional or thick sheath
        print(I_i)
        print(I_i_extras)
        I_i, I_i_extras = get_ion_current(
            shape,
            voltage, current,
            method=sheath_method, V_f=V_f, n_i=n_i, n0=n_e, m_i=m_i, KT_e=KT_e, V_s=V_s,
        )
        # update I_is and I_is extras ## NEEEEEEEEDS ATTENETION
        I_is_extras['fit'] = I_i_extras["fit"]
        print(I_i)
        print(I_i_extras)

        new = lambda_De
        if it != 1:
            difference = old-new
            realtive_difference = difference/old
            RMSE = np.sqrt(np.sum(np.power(realtive_difference, 2)))
        else:
            RMSE = float("inf")
        old = new

        convergence = True if RMSE <= convergence_decimal else False

    results["V_f"]["value"] = V_f
    results["V_f"]["other"] = V_f_extras
    results["V_s"]["value"] = V_s
    results["V_s"]["other"] = V_s_extras
    results["I_is"]["value"] = I_is
    results["I_is"]["other"] = I_is_extras
    results["n_i"]["value"] = n_i
    results["n_i"]["other"] = n_i_extras
    results["I_es"]["value"] = I_es
    results["I_es"]["other"] = I_es_extras
    results["KT_e"]["value"] = KT_e
    results["KT_e"]["other"] = KT_e_extras
    results["n_e"]["value"] = n_e
    results["n_e"]["other"] = n_e_extras
    results["lambda_De"]["value"] = lambda_De
    results["lambda_De"]["other"] = lambda_De_extras
    # results["r_p/lambda_De"]["value"] = ratio
    # results["r_p/lambda_De"]["other"] = ratio_extras
    results["sheath"]["value"] = ratio
    results["sheath"]["other"] = ratio_extras
    data = {'voltage': voltage, 'current': current,
            'current_e': I_e, 'current_i': I_i}
    data = pd.DataFrame(data)
    return data, results



import os

import pyvisa as visa

class TDS640a:
    
    def __init__(self, port = None, mydir = None, myfile = None, channel = None):
         
        # indicates which backend is being used
        rm = visa.ResourceManager()
        print('\n')
        print(rm)

        a = 0
        if port is None:
            a = a + 1
        if channel is None:
            a = a + 1
        if myfile is None:
            a = a + 1
        if mydir is None:
            a = a + 1

        if a != 0:
            print('Not enough arguments passed\nPlease follow prompt\n')
            self.enter()
        else:
            if isinstance(port, int):
                print('\nEntered integer for port, assumeing that meant port #')
                print('\nAvailable Ports:\nPort #:\tPort Name')
                print('__________________')
                b = 0
                for aa in rm.list_resources():
                    b = b + 1
                    print('Port ' + str(b) + ':\t' + aa)
                print('__________________')
                port = rm.list_resources()[port-1]
                print('Using ' + str(port))
            self.gpib = rm.open_resource(port)
            
            self.port = port
            self.myfile = myfile
            self.mydir = mydir
            self.channel = channel  

        # print where
        print('\nWorking with ' + self.port)
        print('Working with ' + self.myfile)
        print('Working in ' + self.mydir)
        print('Working with ' + self.channel)

        print('\n')
        print('General (Floppy and computer) Available  commands:')
        print('changegpibport, changemydir, changemyfile, changechannel')
        print('\n')
        print('Other (Floppy) Available  commands:')
        print('save, delete, cwd, cwdQ, mkdir, rmdir, dirQ ')
        print('\n')
        print('Suggested (computer) Available  commands:')
        print('acquire, saveplt, getwfm ')


    def gpibport(self, PORT = None):
        #import pyvisa as visa 
        # indicates which backend is being used
        rm = visa.ResourceManager()
        try:
            if PORT is None:
                PORT = self.port
            # opens declared resource
            myinstrument = rm.open_resource(PORT)
            return myinstrument
        except IndexError:
            return 0

    def save(self, CHANNEL = None, PORT = None, MYFILE = None):
            
        # NOTES for GPIB TDS640a
        #<filename> = filename of up to 8 characters and can
        #be followed by a period (“.”) and a 3-character extension.
        try:
            if PORT is None:
                PORT = self.port
            if CHANNEL is None:
                CHANNEL = self.channel
            if MYFILE is None:
                MYFILE = self.myfile
            # SAVES WAVEFORM
            # SAVE:WAVEFORM <CH#,MATH#>,"FILEPATH/FILENAME.WFM>"
            #	i.e. (current Directory) SAVE:WAVEFORM CH1,"TEKCH1.WFM"
            #	i.e. (path to directory) SAVE:WAVEFORM CH1,"fd0:/LANG/LANG1/TEKCH1.WFM"
            scope = self.gpibport(PORT)
            scope.write('SAVE:WAVEFORM ' + CHANNEL + ',"' + MYFILE + '"')
            
        except IndexError:
            return 0

    def delete(self, PORT = None, MYFILE = None):
        try:
            if PORT is None:
                PORT = self.port
            if MYFILE is None:
                MYFILE = self.myfile
            # Delete FILE
            # FILESYSTEM:DELTE "<FilePath>"
            scope = self.gpibport(PORT)
            scope.write('FILESYSTEM:DELETE "' + MYFILE + '"')
            
        except IndexError:
            return 0

    def cwdQ(self, PORT = None):
        try:
            if PORT is None:
                PORT = self.port
            # Inquieres what is working directory
            # FILESYSEM:CWD?
            scope = self.gpibport(PORT)
            out = scope.query('FILESYSTEM:CWD?')
            print(out)
            
        except IndexError:
            return print('ERROR')

    def cwd(self, PORT = None, MYDIR = None):
        try:
            if PORT is None:
                PORT = self.port
            if MYDIR  is None:
                MYDIR = self.mydir
            # Sets workin directory
            # FILESYTEM:CWD <DIRPATH>
            #	i.e. FILESYSTEM:CWD "fd0:/LANG/LANG1"
            scope = self.gpibport(PORT)
            scope.write('FILESYSTEM:CWD "' + MYDIR + '"')
            
        except IndexError:
            return 0

    def mkdir(self, PORT = None, MYDIR = None):
        try:
            if PORT is None:
                PORT = self.port
            if MYDIR  is None:
                MYDIR = self.mydir
            # Creates Directory
            # FILESYSTEM:MKDIR "<DIRPATH>"
            #	i.e. FILESYSTEM:MKDIR "fd0:/LANG/LANG2"
            scope = self.gpibport(PORT)
            scope.write('FILESYSTEM:MKDIR "' + MYDIR + '"')
            
        except IndexError:
            return 0

    def rmdir(self, PORT = None, MYDIR = None):
        try:
            if PORT is None:
                PORT = self.port
            if MYDIR is None:
                MYDIR = self.mydir
            # Remove Directory
            #FILESYSTEM:RMDIR "<DIRPATH>"
            #	i.e. FILESYSTEM:RMDIR "fd0:/LANG/LANG2"
            scope = self.gpibport(PORT)
            scope.write('FILESYSTEM:RMDIR "' + MYDIR + '"')
            
        except IndexError:
            return 0
        
    def dirQ(self, PORT = None):
        try:
            if PORT is None:
                PORT = self.port
            # Queires FILES in Directory
            # FILESYSTEM:DIR?
            scope = self.gpibport(PORT)
            out = scope.query('FILESYSTEM:DIR?')
            print(out)
            
        except IndexError:
            return 0
            
    def changegpibport(self, PORT):
        try:
            # opens declared resource
            self.gpib = rm.open_resource(PORT)
            
        except IndexError:
            return ('No Input for Change')
        
    def changemydir(self, MYDIR):
        try:
            # redefines mydir
            self.mydir = MYDIR
            
        except IndexError:
            return print('No Input for Change')

    def changemyfile(self, MYFILE):
        try:
            # redefines myfile
            self.myfile = MYFILE
            
        except IndexError:
            return print('No Input for Change')

    def changechannel(self, CHANNEL):
        try:
            # redefines channel
            self.channel = CHANNEL
            
        except IndexError:
            return print('No Input for Change')

    def acquire(self, CHANNEL = None, PORT = None):
        try:
            if PORT is None:
                PORT = self.port
            if CHANNEL is None:
                CHANNEL = self.channel
            import pstl_getwaveform as getwfm 
            self.wave = getwfm.TDS640a(CHANNEL, PORT)

        except IndexError:
            return print('ERROR in Acquire')

    def saveplt(self, TITLE = None, xlabel = None, ylabel = None):
        try:
            wave = self.wave
            if TITLE is None:
                TITLE = wave.title
                title = "%s" % (TITLE)
            if xlabel is None:
                xlabel = "Time [%s]" % (wave.xunit)
            
            if ylabel is None:
                ylabel = "Voltage [%s]" % (wave.yunit)
            #
            import matplotlib.pyplot as plt
            # plot 
            fig = plt.figure(1, figsize=(6*3/2, 4*3/2))
            plt.clf()
            plt.plot(wave.time, wave.volts)
        
            plt.xlabel(xlabel)
            plt.ylabel(ylabel)
            plt.title(title)
            #plt.show()
            plt.savefig(self.mydir + wave.name + ".png")
            self.wave.fig = fig
            #plt.close(1)
        except IndexError:
            return 0

    def savewfm(self, MYDIR = None, MYFILE = None):
        try:
            if MYFILE is None:
                MYFILE = self.myfile
            if MYDIR is None:
                MYDIR = self.mydir
            h = self.wave
            aa = ('ymult', 'yzero', 'yoff', 'xincr', 'xdelay', 'volts', 'time', 'xunit',\
             'yunit', 'name', 'title', 'wfid', 'wfmpre')
            bb = (repr(h.ymult), repr(h.yzero), repr(h.yoff), repr(h.xincr), repr(h.xdelay),\
             repr(h.volts), repr(h.time), repr(h.xunit), repr(h.yunit),\
             repr(h.name), repr(h.title), repr(h.wfid), repr(h.wfmpre))

            file = open(MYDIR + MYFILE, "w")
            for a in range(1,len(aa)): 
                file.write(aa[a] + ' = ' + bb[a] + '\n')
            file.close
        except IndexError:
            return 0

    def getwfm(self, CHANNEL = None, PORT = None, TITLE = None, xlabel = None, ylabel = None,\
     MYDIR = None, MYFILE = None):
        try:
            self.acquire(CHANNEL, PORT)
            self.savewfm(MYDIR, MYFILE)
            self.saveplt(TITLE, xlabel, ylabel)
        except IndexError:
            return 0

    def enter(self):
        try:
            #import pyvisa as visa 
            # indicates which backend is being used
            rm = visa.ResourceManager()

            # First time
            print('\nAvailable Ports:(Enter "R" for refresh)\nPort #:\tPort Name')
            print('__________________')
            a = 0
            for aa in rm.list_resources():
                a = a + 1
                print('Port ' + str(a) + ':\t' + aa)
            print('__________________')
            strin = input("Enter Port #:\n>>")
            # end of first

            # if refresh
            while strin == "R":
                print('\nAvailable Ports:(Enter "R" for refresh)\nPort #:\tPort Name')
                print('__________________')
                a = 0
                for aa in rm.list_resources():
                    a = a + 1
                    print('Port ' + str(a) + ':\t' + aa)
                print('__________________')
                strin = input("Enter Port #:\n>>")
            # end of refresh
            port = rm.list_resources()[int(strin)-1]
            print('maybe change later with changegpibport(<newport>)')
            print('port = ' + port)

            print('\nSet Working Directory')
            print('i.e. For floppy fd0:/mydir/, input>> fd0:/mydir/')
            print('i.e. For windows computer C:\\User\\<youruser>\\Downloads\\, input>> ~\\Downloads\\')
            print('\t or input>> C:\\\\User\\\\<youruser>\\\\Downloads\\\\')
            print('maybe change later with changemydir(<newdir>)')
            mydir = input('mydir =\n>>')
            if '~\\' in mydir:
                mydir = os.path.expanduser(mydir)
            print('mydir = ' + mydir)

            print('\nSet default save file')
            print('i.e. TEK0.WFM')
            print('maybe change later with changemyfile(<newfile>)')
            myfile = input('myfile =\n>>')
            print('myfile = ' + myfile)

            print('\nSet default Channel, Math, or Reference')
            print('i.e. CH2, MATH1, REF3')
            print('maybe change later with changechannel(<newchannel>)')
            channel = input('channel =\n>>')
            print('channel = ' + channel)

            self.gpib = rm.open_resource(port)
        
            self.port = port
            self.myfile = myfile
            self.mydir = mydir
            self.channel = channel

        except IndexError:
            return 0



"""

pstl-tools


Tools for working with PSTL equipment easier

"""

__version__ = '0.1.0'
__author__ = 'tyjoto'

from pstl import diagnostics
from pstl import gui
from pstl import instruments
from pstl import scripts
from pstl import utls

__all__ = [
    diagnostics,
    gui,
    instruments,
    scripts,
    utls,
]   # type: ignore



import numpy as np

from pstl.diagnostics.probes.langmuir.single.analysis.floating_potential import check_for_floating_potential, get_below_floating_potential
from pstl.utls import constants as c
from pstl.utls.helpers import method_function_combiner
from pstl.utls.helpers import method_selector as method_selector
from pstl.utls.helpers import find_fit
from pstl.utls.helpers import FunctionFit


default_fit_linear_kwargs = {
    'deg': 1, 'power': 1, 'polarity': 1,
    'reverse': False, 'return_best': True, 'fit_type': "linear",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': None, 'fstep_type': None, 'fstep_adjust': True, 'fitmax': None,
    'bstep': None, 'bstep_type': None, 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.30,
    'threshold_rsq': 0.85, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 10.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}

default_fit_power_kwargs = {
    'deg': 1, 'power': None, 'polarity': -1,  # power needs to be defined
    'reverse': False, 'return_best': True, 'fit_type': "power",
    'min_points': 5, 'istart': None, 'iend': None, 'invalid': "ignore",
    'fstep': None, 'fstep_type': None, 'fstep_adjust': True, 'fitmax': None,
    'bstep': None, 'bstep_type': None, 'bstep_adjust': True, 'bitmax': None,
    'threshold_residual': None, 'threshold_rmse': 0.50,
    'threshold_rsq': 0.95, 'threshold_method': None,
    'convergence_residual_percent': 1.0, 'convergence_rmse_percent': 10.0,
    'convergence_rsq_percent': 1.0, 'convergence_method': None,
    'strict': False, 'full': True, 'printlog': False,
}


def shape_fit_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_methods, available_fit_functions)
    return func


def shape_func_selector(shape):
    # select function based on integer_method
    func = method_selector(shape, available_methods, available_func_functions)
    return func


def cylinderical_fit_method(voltage, current, *args, **kwargs):
    # check  and setup xdata,ydata based on floaing potential
    V_f = kwargs.pop("V_f", None)
    V_f = check_for_floating_potential(V_f, voltage, current, **kwargs)
    iend, xdata, ydata = get_below_floating_potential(V_f, voltage, current)

    # I_i function
    fit_kwargs = dict(default_fit_power_kwargs)
    fit_kwargs["power"] = 0.75
    fit_kwargs.update(kwargs.pop("fit_kwargs", {}))
    fit = find_fit(xdata, ydata, **fit_kwargs)
    # silence error when computing
    with np.errstate(invalid="ignore"):
        I_i = fit(voltage)

    # condition for I_i is negative
    condition = I_i <= 0

    # get I_i
    I_i = np.where(
        condition,
        I_i,
        0,
    )

    # make returns
    extras = {"fit": fit, "method": "thin", "shape": "cylinderical"}
    return I_i, extras


def spherical_fit_method(voltage, current, *args, **kwargs):

    # check  and setup xdata,ydata based on floaing potential
    V_f = kwargs.pop("V_f", None)
    V_f = check_for_floating_potential(V_f, voltage, current, **kwargs)
    iend, xdata, ydata = get_below_floating_potential(V_f, voltage, current)

    # I_i function
    fit_kwargs = dict(default_fit_power_kwargs)
    fit_kwargs["power"] = 1.5
    fit_kwargs.update(kwargs.pop("fit_kwargs", {}))
    fit = find_fit(xdata, ydata, **fit_kwargs)
    # silence error when computing
    with np.errstate(invalid="ignore"):
        I_i = fit(voltage)

    # condition for I_i is negative
    condition = I_i <= 0

    # get I_i
    I_i = np.where(
        condition,
        I_i,
        0,
    )

    # make returns
    extras = {"fit": fit, "method": "thin", "shape": "spherical"}
    return I_i, extras


def planar_fit_method(voltage, current, *args, **kwargs):
    # check  and setup xdata,ydata based on floaing potential
    V_f = kwargs.pop("V_f", None)
    V_f = check_for_floating_potential(V_f, voltage, current, **kwargs)
    iend, xdata, ydata = get_below_floating_potential(V_f, voltage, current)

    # I_i function
    fit_kwargs = dict(default_fit_linear_kwargs)
    fit_kwargs.update(kwargs.pop("fit_kwargs", {}))
    fit = find_fit(xdata, ydata, **fit_kwargs)
    # silence error when computing
    with np.errstate(invalid="ignore"):
        I_i = fit(voltage)

    # condition for I_i is negative
    condition = I_i <= 0

    # get I_i
    I_i = np.where(
        condition,
        I_i,
        0,
    )

    # make returns
    extras = {"fit": fit, "method": "thin", "shape": "planar"}
    return I_i, extras


def linear_fit_method(voltage, current, *args, **kwargs):
    # check  and setup xdata,ydata based on floaing potential
    V_f = kwargs.get("V_f", None)
    V_f = check_for_floating_potential(V_f, voltage, current, **kwargs)
    iend, xdata, ydata = get_below_floating_potential(V_f, voltage, current)

    # I_i function
    fit_kwargs = dict(default_fit_linear_kwargs)
    fit_kwargs.update(kwargs.pop("fit_kwargs", {}))
    fit = find_fit(xdata, ydata, **fit_kwargs)
    # silence error when computing
    with np.errstate(invalid="ignore"):
        I_i = fit(voltage)

    # condition for I_i is negative
    condition = I_i <= 0

    # get I_i
    I_i = np.where(
        condition,
        I_i,
        0,
    )

    # make returns
    extras = {"fit": fit, "method": "thin-linear", "shape": "None"}
    return I_i, extras


def cylinderical_func_method(voltage, area, n0, V_s, m_i, KT_e, r_p=None, lambda_D=None, *args, correct_area=True, **kwargs):
    if correct_area is True:
        if r_p is None or lambda_D is None:
            raise ValueError("'r_p' and 'lambda_D' must be a float: " +
                             str(type(r_p))+" and "+str(type(lambda_D)))
        xs = _xs(voltage, V_s, KT_e, lambda_D)
        area = area*(1+xs/r_p)
    else:
        area = np.ones_like(voltage)*area

    # condition
    condition = cylinderical_domain_condition(voltage, V_s, KT_e)

    # I_i function
    func = _function

    # get I_i
    I_i = np.where(
        condition,
        func(area, n0, m_i, KT_e),
        0,
    )

    # make returns
    extras = {"method": "thin", "shape": "cylinderical"}
    return I_i, extras


def spherical_func_method(voltage, area, n0, V_s, m_i, KT_e, r_p=None, lambda_D=None, *args, correct_area=True, **kwargs):
    if correct_area is True:
        if r_p is None or lambda_D is None:
            raise ValueError("'r_p' and 'lambda_D' must be a float: " +
                             str(type(r_p))+" and "+str(type(lambda_D)))
        xs = _xs(voltage, V_s, KT_e, lambda_D)
        area = area*np.power(1+xs/r_p, 2)
    else:
        area = np.ones_like(voltage)*area

    # condition
    condition = spherical_domain_condition(voltage, V_s, KT_e)

    # I_i function
    func = _function

    # get I_i
    I_i = np.where(
        condition,
        func(area, n0, m_i, KT_e),
        0,
    )

    # make returns
    extras = {"method": "thin", "shape": "spherical"}
    return I_i, extras


def planar_func_method(voltage, area, n0, V_s, m_i, KT_e, r_p=None, lambda_D=None, *args, correct_area=True, **kwargs):
    if correct_area is True:
        pass  # no correction if oriented correctly
        if r_p is None or lambda_D is None:
            # not needed because no correction (filler for later if orientation bad)
            pass
        area = np.ones_like(voltage)*area
    else:
        area = np.ones_like(voltage)*area

    # condition
    condition = planar_domain_condition(voltage, V_s, KT_e)

    # I_i function
    func = _function

    # get I_i
    I_i = np.where(
        condition,
        func(area, n0, m_i, KT_e),
        0,
    )

    # make returns
    extras = {"method": "thin", "shape": "planar"}
    return I_i, extras
# domain conditions


def cylinderical_domain_condition(voltage, V_s, KT_e):
    return _domain_condition(voltage, V_s, KT_e)


def spherical_domain_condition(voltage, V_s, KT_e):
    return _domain_condition(voltage, V_s, KT_e)


def planar_domain_condition(voltage, V_s, KT_e):
    return _domain_condition(voltage, V_s, KT_e)

# generalized domain


def _domain_condition(voltage, V_s, KT_e):
    condition = np.divide(
        np.subtract(V_s, voltage),  # [V]
        KT_e                         # [eV]
    )
    return condition > 1  # >>1


def _xs(voltage, V_s, KT_e, lambda_D):
    A = lambda_D*np.sqrt(2)/3
    x = 2*(V_s-voltage)/KT_e
    with np.errstate(invalid="ignore"):
        power_x = np.power(x, 0.75)
    return np.multiply(A, power_x)


def _function(area, n0, m_i, KT_e):
    return np.exp(-0.5)*c.e*n0*area*np.sqrt(c.e*KT_e/m_i)


def _constant_function(voltage, current, *args, **kwargs):
    return np.ones_like(voltage)*_function(*args)


# Declare available methods
available_methods = {
    0: ['cylinderical'],
    1: ['spherical'],
    2: ['planar', 'planer'],
}
# declare correspondin functions for available_methods
available_fit_functions = {
    0: cylinderical_fit_method,
    1: spherical_fit_method,
    2: planar_fit_method,
}
# declare correspondin functions for available_methods
available_func_functions = {
    0: cylinderical_func_method,
    1: spherical_func_method,
    2: planar_func_method,
}

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
fit_options = method_function_combiner(
    available_methods, available_fit_functions)

# combines available_methods and available_methods and must have the same keys
# such that the new dictionary has the same keys but value is a tuple(list(str),functions)
func_options = method_function_combiner(
    available_methods, available_func_functions)



from pstl.utls.plasmas.sheath import transitional
from pstl.utls.plasmas.sheath import thick

__all__ = [
    'find_ab_coefs',
    transitional,
    thick,
]


def find_ab_coefs(sheath, shape, *args, **kwargs):
    raise ValueError("find ab coefs")


def get_probe_to_sheath_ratio(r_p, lambda_D, *args, **kwargs):
    ratio = r_p/lambda_D
    if ratio <= 3:
        sheath_type = "thick"
    elif ratio > 3 and ratio < 50:
        sheath_type = "transitional"
    elif ratio >= 50:
        sheath_type = "thin"
    else:
        raise ValueError(
            "ratio is not in bounds check r_p and lambda_D: {r_p} and {lambda_D}")
    return ratio, {"sheath": sheath_type}



class FitConvergenceError(Exception):
    """Raised when Convergence of fit routine could not be found. 
    This was created so that a try and except could be used to account for convergence error such that 
    another method maybe used to achieve same result as this fit routine."""


class MissingReturnError(Exception):
    """Raised when a return argument was not suppose to be empty, but was being returned empty. 
    This is most likely a bug in the code"""

    def __init__(self, msg="Missing Return argument. Maybe a bug in the code"):
        self.msg = msg
        Exception.__init__(self, self.msg)




from collections.abc import Sequence
import tkinter as tk
import numpy as np
import pandas as pd
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2Tk  # type: ignore
from scipy.signal import find_peaks, savgol_filter
from scipy.optimize import curve_fit

from pstl.utls.verify import verify_pair_of_1D_arrays, verify_type
from pstl.utls.helpers import normalize
from pstl.utls.helpers import _fit_exponential_func, _fit_linear_func, _fit_polynomial_func
from pstl.utls.helpers_savgol import set_savgol_filter_defaults


def smooth_filter(data, polyorder=1, window_length=1, *args, **kwargs):
    voltage = data.voltage
    current = data.current

    # perform preprocessing filter
    current = savgol_filter(
        current, polyorder=polyorder, window_length=window_length, *args, **kwargs)

    smoothed_data = pd.DataFrame({"voltage": voltage, "current": current})

    return smoothed_data


def preprocess_filter(data):
    voltage = data.voltage
    current = data.current

    # perform preprocessing filter
    voltage, current, pts_deleted = preprocessing_filter(
        voltage, current, delete=True, interactive=False,
    )

    filtered_data = pd.DataFrame({"voltage": voltage, "current": current})
    deleted_data = pd.DataFrame(
        {
            "voltage": pts_deleted[:, 0], "current": pts_deleted[:, 1]
        }
    )
    return filtered_data, deleted_data


def preprocessing_filter(xdata: np.ndarray, ydata: np.ndarray,
                         delete: bool = True, replace: bool = False, interp: str = 'linear',
                         interactive: bool = False,
                         *args, **kwargs) -> tuple[np.ndarray, np.ndarray, np.ndarray | None]:
    """
    Preprocesses xdata and ydata by applying Savitzky-Golay filter for smoothing the curve, 
    detecting erroneous points using find_peaks, and optionally replacing erroneous points with 
    interpolated values. 

    Parameters:
    -----------
    xdata: np.ndarray
        Array containing x-coordinates of the data.
    ydata: np.ndarray
        Array containing y-coordinates of the data.
    delete: bool, default True
        Boolean indicating whether to delete the erroneous points or not. If True, deletes the points.
    replace: bool, default False
        Boolean indicating whether to replace the erroneous points with interpolated values or not.
    interp: str, default 'linear'
        String indicating the interpolation scheme for replacing the erroneous points. 
        Must be one of 'linear', 'exponential', 'polynomial'.
    interactive: bool, default False
        Boolean indicating whether to show interactive matplotlib to get better filtering.
    **kwargs:
        Additional keyword arguments that can be passed to savgol_filter or find_peaks functions.
        The keyword arguments **kwargs are used to specify additional options for the Savitzky-Golay filter and the find_peaks function. The possible keyword arguments are:
        - savgol_filter_kw: A dictionary of keyword arguments for the Savitzky-Golay filter. The default value is {}.
            - window_length: int, optional
                Length of the filter window. If None, it is calculated as 0.08 * len(ydata). Default is None.
            - polyorder: int, optional
                Order of the polynomial used to fit the samples. Default is 1.
        - find_peaks_kw: A dictionary of keyword arguments for the find_peaks function. The default value is {}.
            - prominence: float, optional
                Minimum prominence of peaks. Normalized between [0, 1]. Default is 0.5.
            - threshold: float, optional
                Minimum height of peaks. Normalized between [0, 1]. Default is 0.2.

    Returns:
    --------
    Tuple[np.ndarray, np.ndarray, np.ndarray | None]
        Tuple containing the preprocessed xdata and ydata and x,y points modified.

        If delete=True and replace=False: A tuple of two numpy arrays (xdata, ydata) with erroneous points removed.
        If delete=False and replace=True: A tuple of two numpy arrays (xdata, ydata) with erroneous points replaced by interpolated values.
        If delete=True and replace=True: A tuple of two numpy arrays (xdata, ydata) with erroneous points removed and then replaced by interpolated values.
    """
    # Verify Arrays
    # check if xdata and ydata are np.ndarray, if not convert them to np.ndarray
    # and get shape/lengths of data and verify they are the same.
    xdata, ydata = verify_pair_of_1D_arrays(xdata, ydata)  # type: ignore
    len_data = len(xdata)

    # Get Defaults from keywords for savgol filter and findpeaks
    # retrive savgol_filter and find_peaks keyword arguments from kwargs
    savgol_filter_kw = kwargs.get('savgol_filter_kw', {})
    find_peaks_kw = kwargs.get('find_peaks_kw', {})

    # Savgol_filter for smoothing curve to find difference to identify
    # errounous (spikes) in data

    # Set defaults for savgol_filter
    window_length, polyorder, savgol_filter_kw = set_savgol_filter_defaults(
        ydata, savgol_filter_kwargs=savgol_filter_kw)

    # apply savgol fiter to ydata
    ydata_filtered = savgol_filter(
        ydata, window_length, polyorder, **savgol_filter_kw)

    # Determine where errounous points (spikes in data) occur using find_peaks
    # find difference of ydata and savgol_filtered data, then absolute value and
    # normalize
    ydata_abs_diff = np.abs(np.subtract(ydata, ydata_filtered))
    ydata_normalized = normalize(ydata_abs_diff)

    # get defaults for find_peaks if none are given
    find_peaks_kw['prominence'] = find_peaks_kw.get('prominence', 0.5)
    find_peaks_kw['threshold'] = find_peaks_kw.get('threshold', 0.2)

    # verify is either float, integer or None, raise error if they are not
    if isinstance(find_peaks_kw['prominence'], (int, float)) is False:
        if find_peaks_kw['prominence'] is not None:
            raise TypeError("'prominence' is not an integer or float: %s" % (
                str(find_peaks_kw['prominence'])))
    if isinstance(find_peaks_kw['threshold'], (int, float)) is False:
        if find_peaks_kw['prominence'] is not None:
            raise TypeError("'threshold' is not an integer or float: %s" % (
                str(find_peaks_kw['threshold'])))
    # verify the values are less then one as these are normalized peaks
    if find_peaks_kw['prominence'] > 1:
        if find_peaks_kw['prominence'] is not None:
            raise ValueError("peaks are normalized, 'prominence' must be a value less than 1 or None: %s" % (
                str(find_peaks_kw['prominence'])))
    if find_peaks_kw['threshold'] > 1:
        if find_peaks_kw['threshold'] is not None:
            raise ValueError("peaks are normalized, 'threshold' must be a value less than 1 or None: %s" % (
                str(find_peaks_kw['threshold'])))

    # solve and identify peaks
    peaks, props = find_peaks(ydata_normalized, **find_peaks_kw)

    # if interp is 'linear', 'exponetial', 'polynomial.
    interp_strings = ['linear', 'exponential', 'polynomial']
    if replace and interp in interp_strings:
        if interp in ['linear']:
            fit_func = _fit_linear_func
        elif interp in ['exponential']:
            fit_func = _fit_exponential_func
        elif interp in ['polynomial']:
            fit_func = _fit_polynomial_func
        else:
            raise ValueError("wrong interpolation scheme")
        xdata_points_to_replace = xdata[peaks]
        ydata_points_to_replace = ydata[peaks]
        # loop through indeces to replace
        ydata_modified = ydata.copy()
        xdata_modified = xdata.copy()
        for ipeak in peaks:
            if ipeak != 0 and ipeak != len_data:
                ilower = ipeak-1
                ihigher = ipeak+1
            elif ipeak == 0:
                ilower = ipeak+1
                ihigher = ipeak+2
            elif ipeak == len_data:
                ilower = ipeak-2
                ihigher = ipeak+1
            else:
                raise ValueError(
                    "Something went wrong with indecies during replacing points")
            popt, pcov = curve_fit(fit_func,
                                   xdata_points_to_replace[ilower, ihigher],
                                   ydata_points_to_replace[ilower, ihigher],
                                   )
            ydata_fitted = fit_func(xdata_points_to_replace,
                                    *tuple(popt))
            ydata_modified[ipeak] = ydata_fitted
        points_modified = np.array([xdata[peaks], ydata[peaks]]).transpose()
    # if delete is True
    elif delete:
        xdata_modified = np.delete(xdata, peaks)
        ydata_modified = np.delete(ydata, peaks)
        points_modified = np.array([xdata[peaks], ydata[peaks]]).transpose()
    # if replace but wrong interp
    elif replace is False or interp not in interp_strings:
        raise ValueError("Error in 'replace is %s and 'interp' is %s" %
                         (str(replace), str(interp)))
    else:
        xdata_modified = xdata.copy()
        ydata_modified = ydata.copy()
        points_modified = None

    # if interactive plot, makes plots
    if interactive:
        # create root tkinter window to house the plots
        root = tk.Tk()
        # create frames to store figure canvas (draw on)
        frame_figure = tk.Frame(root)
        # and the create frame for control panel packed next to canvas
        frame_control_panel = tk.Frame(root)
        # create control panels for filter and peaks
        frame_cp_filter = tk.Frame(frame_control_panel)
        frame_cp_peaks = tk.Frame(frame_control_panel)
        # create figure to be added to the window
        fig = Figure(figsize=(8, 6), dpi=100)
        # autoadjust the layout to tight
        fig.set_tight_layout(True)
        # initialize figure
        # A tk.DrawingArea.
        canvas = FigureCanvasTkAgg(fig, master=frame_figure)
        canvas.draw()
        wgt_canvas = canvas.get_tk_widget()

        # pack_toolbar=False will make it easier to use a layout manager later on.
        wgt_toolbar = NavigationToolbar2Tk(
            canvas, frame_figure, pack_toolbar=False)
        wgt_toolbar.update()

        # get the axes object
        ax_filter = fig.add_subplot(211)  # type: ignore
        ax_peaks = fig.add_subplot(212, sharex=ax_filter)  # type: ignore
        # set title
        # fig.title("Filtering Erronous Points")
        # add raw data plot
        raw_data_kwargs = kwargs.get('raw_data_kwargs', {})
        raw_data_kwargs.setdefault('color', 'C0')
        raw_data_kwargs.setdefault('alpha', 0.3)
        raw_data_kwargs.setdefault('markerfacecolor', 'none')
        raw_data_kwargs.setdefault('label', 'Prefiltered Data')
        raw_data_kwargs.setdefault('linestyle', '-')
        raw_data_kwargs.setdefault('marker', '*')
        line_raw_data, = ax_filter.plot(  # type: ignore
            xdata, ydata, **raw_data_kwargs)  # type: ignore

        # add filterd data plot
        filtered_data_kwargs = kwargs.get('filtered_data_kwargs', {})
        filtered_data_kwargs.setdefault('color', 'C1')
        filtered_data_kwargs.setdefault('alpha', 0.5)
        filtered_data_kwargs.setdefault('markerfacecolor', 'none')
        filtered_data_kwargs.setdefault('label', 'Filtered Data')
        filtered_data_kwargs.setdefault('linestyle', '-')
        filtered_data_kwargs.setdefault('marker', '^')
        line_filtered_data, = ax_filter.plot(  # type: ignore
            xdata, ydata_filtered, **filtered_data_kwargs)

        # add modified returning data
        modified_data_kwargs = kwargs.get('modified_data_kwargs', {})
        modified_data_kwargs.setdefault('color', 'C2')
        modified_data_kwargs.setdefault('alpha', 1)
        modified_data_kwargs.setdefault('markerfacecolor', 'none')
        modified_data_kwargs.setdefault('label', 'Modified Data')
        modified_data_kwargs.setdefault('linestyle', '-')
        modified_data_kwargs.setdefault('marker', 'v')
        line_modified_data, = ax_filter.plot(  # type: ignore
            xdata_modified, ydata_modified, **modified_data_kwargs)

        # points deleted
        deleted_data_kwargs = kwargs.get('deleted_data_kwargs', {})
        deleted_data_kwargs.setdefault('color', 'C3')
        deleted_data_kwargs.setdefault('alpha', 1)
        # deleted_data_kwargs.setdefault('markerfacecolor', 'None')
        deleted_data_kwargs.setdefault('label', 'Erroneous Spikes')
        deleted_data_kwargs.setdefault('linestyle', 'None')
        deleted_data_kwargs.setdefault('marker', 'X')
        line_peaks, = ax_filter.plot(
            xdata[peaks], ydata[peaks], **deleted_data_kwargs
        )

        # add nomralized, abs residual and identified peaks
        residual_plot_kwargs = kwargs.get('residual_plot_kwargs', {})
        residual_plot_kwargs.setdefault('color', 'C2')
        residual_plot_kwargs.setdefault('alpha', 1)
        residual_plot_kwargs.setdefault('markerfacecolor', 'none')
        residual_plot_kwargs.setdefault(
            'label', 'Normalized Absolute Resdiual')
        residual_plot_kwargs.setdefault('linestyle', '-')
        residual_plot_kwargs.setdefault('marker', 'o')
        line_residual, = ax_peaks.plot(
            xdata, ydata_normalized, **residual_plot_kwargs
        )

        # add nomralized, abs residual and identified peaks
        peaks_plot_kwargs = kwargs.get('peaks_plot_kwargs', {})
        peaks_plot_kwargs.setdefault('color', 'C3')
        peaks_plot_kwargs.setdefault('alpha', 1)
        # peaks_plot_kwargs.setdefault('markerfacecolor', 'None')
        peaks_plot_kwargs.setdefault('label', 'Erroneous Spikes')
        peaks_plot_kwargs.setdefault('linestyle', 'None')
        peaks_plot_kwargs.setdefault('marker', 'X')
        line_peaks, = ax_peaks.plot(
            xdata[peaks], ydata_normalized[peaks], **peaks_plot_kwargs
        )

        # turn legends on
        ax_filter.legend()
        ax_peaks.legend()

        # Declare the variabels for tkinter widgets
        var_window_length = tk.IntVar()
        var_polyorder = tk.IntVar()
        # set default values
        var_window_length.set(window_length)
        var_polyorder.set(polyorder)

        def update_filter(window_length: int, polyorder: int, ydata: Sequence[float],
                          line_filtered_data: Line2D, **savgol_filter_kw) -> None:
            # apply filter with updated parameters
            ydata_filtered = savgol_filter(
                ydata, window_length, polyorder, **savgol_filter_kw)

            # set the filtered data to the updated values
            line_filtered_data.set_ydata(ydata_filtered)
            canvas.draw()

        # Make a horizontal slider to control the frequency.
        window_length_slider = tk.Scale(
            frame_cp_filter, variable=var_window_length,  # type: ignore
            from_=2, to=len_data, label="Window length",
            command=lambda val: update_filter(var_window_length.get(),
                                              var_polyorder.get(),
                                              line_raw_data.get_ydata(),
                                              line_filtered_data,
                                              **savgol_filter_kw)
        )

        # Make a vertically oriented slider to control the amplitude
        polyorder_slider = tk.Scale(
            frame_cp_filter, variable=var_polyorder,  # type: ignore
            from_=1, to=len_data-1, label="Polyorder",
            command=lambda val: update_filter(
                var_window_length.get(),
                var_polyorder.get(),
                line_raw_data.get_ydata(),
                line_filtered_data,
                **savgol_filter_kw)
        )

        # Create a `matplotlib.widgets.Button` to reset the sliders to initial values.
        def reset_filter(window_length: int, polyorder: int, ydata: Sequence[float],
                         line_filtered_data: Line2D, **savgol_filter_kw) -> None:
            window_length_slider.set(window_length)
            polyorder_slider.set(polyorder)
            update_filter(
                int(window_length_slider.get()),
                int(polyorder_slider.get()),
                ydata,
                line_filtered_data,
                **savgol_filter_kw
            )
        btn_filter_reset = tk.Button(
            frame_cp_filter, text="Reset",  # type: ignore
            command=lambda: reset_filter(
                window_length,
                polyorder,
                line_raw_data.get_ydata(),
                line_filtered_data,
                **savgol_filter_kw)
        )

        # packing filter widgets
        window_length_slider.grid(  # type: ignore
            row=0, column=0, sticky="NWSE")
        polyorder_slider.grid(  # type: ignore
            row=0, column=1, sticky="NWSE")
        btn_filter_reset.grid(  # type: ignore
            row=1, column=0, columnspan=2, sticky="NWSE")

        # pack in canvas and toolbar widgets
        # wgt_canvas.grid(  # type: ignore
        # row=0, column=0, sticky="NSWE")
        # wgt_toolbar.grid(  # type: ignore
        # row=1, column=0, sticky="NSWE")
        wgt_toolbar.pack(side=tk.BOTTOM, fill=tk.X)  # type: ignore
        wgt_canvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)  # type: ignore

        # pack in sub control panel frames to main control panel frame
        frame_cp_filter.grid(  # type:ignore
            row=0, column=0, sticky="NWSE")
        frame_cp_peaks.grid(  # type:ignore
            row=1, column=0, sticky="NWSE")

        # pack in control panel to left of canvas
        frame_control_panel.grid(  # type: ignore
            row=0, column=0, sticky="NEWS")
        frame_figure.grid(  # type: ignore
            row=0, column=1, sticky="NEWS")

        # run mainloop
        root.mainloop()  # type: ignore
    elif interactive is False:
        pass
    else:
        raise ValueError("'interactive' only accepts bool: '%s' and '%s'" % (
            str(interactive)))

    # return tulpe of modified xdata
    return xdata_modified, ydata_modified, points_modified



from typing import Any

import numpy as np

from pstl.utls.protocol.gpib.pyvisa import Open
from pstl.instruments.daq.agilent import cards
from pstl.instruments.daq.agilent import commands as cmds

class Agilent34970A(Open):
    def __init__(self,port=None):

        Open.__init__(self,port)

        self.class_name="agilent34970A"
        self.type="daq"

        self.name=self.query("*IDN?")

        card : list[Any] = [None]*4    # three potential cards and 0 is list cards
        card[0]=self.list_cards
        self.card=card
    
    def getVDC(self,loc):
        return self.query(cmds.cmdGetVoltageDC(loc))

    def getVAC(self,loc):
        return self.query(cmds.cmdGetVoltageAC(loc))

    def getTempTCK(self,loc):
        return self.query(cmds.cmdGetTemperatureTCK(loc))

    def getRes(self,loc):
        return self.query(cmds.cmdGetResistance(loc))


    def addCard(self):
        # add interative later
        pass


    def addCardAgilent34901A(self,slot,nchannels,chtype):

        self.card[slot]=\
        cards.Agilent34901A(slot,nchannels,chtype)


    def list_cards(self):
        card=self.card
        print()
        for k in range(1,len(card)):
            try:
                print("Slot: %s\nType: %s\nChannels: %s\n"%(str(k),str(card[k].name),str(card[k].nchannels)))
                self.card[k].channel[0]()
            except:
                print("Slot: %s\nType: None\n\n"%(str(k)))

    def get(self,location,channel=None):
        """
        Gets the preset value for that slot/channel combo
        if only location then 'slot#,channel##'
        if both location and channel then location is slot
        """
        if channel is None:
            loc=location
            location=int(np.floor(loc/100))
            channel = int(loc-location*100)
        cmd=self.card[location].channel[channel].getcmd
        return self.query(cmd)



import string
from typing import Any, Optional, Union, Tuple, Sequence, Callable

import numpy as np
import numpy.typing as npt
from numpy.polynomial import Polynomial as P
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from scipy.signal import savgol_filter, find_peaks
import pandas as pd

from pstl.utls.errors import FitConvergenceError, MissingReturnError
from pstl.utls.verify import verify_iterables, verify_1D_array, verify_type


class FunctionFit:
    def __init__(
            self,
            func: Callable[[npt.ArrayLike, tuple[int | float, ...]], int | float],
            coef: Sequence,
            poly: P,
            domain: Sequence,
            stats: dict,
    ) -> None:

        self.func = func
        self.coef = coef
        self.poly = poly
        self.domain = domain
        self.stats = stats

    def __call__(self, x: npt.ArrayLike, *args: Any, **kwds: Any) -> Any:
        return self.func(x, *self.coef)

    def xrange(self, n: int | None = None, domain: Optional[Sequence] = None):
        if domain is None:
            domain = self.domain  # type: ignore
        if n is None:
            n = int((domain[1]-domain[0])/1e-3)

        return np.linspace(domain[0], domain[1], n)  # type: ignore

    def yrange(self, n: int | None = None, domain: Optional[Sequence] = None):
        if domain is None:
            domain = self.domain  # type: ignore
        if n is None:
            n = int((domain[1]-domain[0])/1e-3)

        # type: ignore
        return self.__call__(np.linspace(domain[0], domain[1], n))


class ExponentialFit(FunctionFit):
    def __init__(self, poly: P, domain: npt.ArrayLike, stats: dict = {}) -> None:
        func = _fit_exponential_func
        coef = poly.convert().coef
        super().__init__(func, coef, poly, domain, stats)  # type: ignore


class PolynomialFit(FunctionFit):
    def __init__(self, poly: P, domain: npt.ArrayLike, stats: dict = {}) -> None:
        func = _fit_polynomial_func
        coef = poly.convert().coef
        super().__init__(func, coef, poly, domain, stats)  # type: ignore


class PowerFit(FunctionFit):
    def __init__(self, power: int | float, poly: P, domain: Sequence, stats: dict = {}, polarity: int = 1) -> None:
        func = _fit_power_func
        coef = poly.convert().coef
        self.power = power
        self._polarity = polarity
        super().__init__(func, coef, poly, domain, stats)

    def __call__(self, x: npt.ArrayLike, *args: Any, **kwds: Any) -> Any:
        result = self.func(
            self.power,
            x,  # type: ignore
            *self.coef)*self._polarity
        return result


class CustomFit(FunctionFit):
    def __init__(self, func, coef, domain, fit_xrange, stats={}):
        poly = P([0, 1], domain=fit_xrange)
        super().__init__(func, coef, poly, domain, stats)  # type: ignore


def make_CustomFit(func, voltage, current, coefs):
    domain = [np.min(voltage), np.max(voltage)]
    # make fit function
    indexs = np.where(current != 0)[0]
    vmin = voltage[indexs[0]]
    vmax = voltage[indexs[-1]]
    fit_xrange = [vmin, vmax]

    fit = CustomFit(func, coefs, domain, fit_xrange)
    return fit


def safe_sqrt(x):
    mask = x < 0
    sqrt_x = np.sqrt(np.abs(x))
    sqrt_x[mask] = np.nan
    return sqrt_x


def method_not_found(method, available_methods):

    # make table string to print
    table = "\n".join([f"{k}\t{v}" for k, v in available_methods.items()])

    # raise error and make table
    raise ValueError(
        f"Matching method not found: {method}\nChoose from one of the available options:\n{table}")


def get_method_options(method, available_methods) -> tuple[int, str]:
    imethod: int | None = None
    vmethod: str | None = None
    if isinstance(method, int) and method in available_methods:
        # perform action using list of strings tied to the integer key
        imethod = method
        vmethod = available_methods[method][0]

    elif isinstance(method, str):
        # perform action using the string as the method name
        # check if the string matches any of the available_methods
        for key, value in available_methods.items():
            if method.lower() in value:
                imethod = key
                vmethod = value[0]
    else:
        imethod = None
        vmethod = None

    # make returns
    if imethod is None or vmethod is None:
        returns: tuple[int, str] = (0, "")
        # will raise error below
        method_not_found(method, available_methods)
    else:
        returns: tuple[int, str] = (imethod, vmethod)

    return returns


def method_selector(method, available_methods, available_functions):
    imethod, _ = get_method_options(method, available_methods)
    func = available_functions[imethod]
    return func


def method_function_combiner(dict1, dict2):
    if set(dict1.keys()) != set(dict2.keys()):
        raise ValueError("The two dictionaries do not have the same keys!")
    new_dict = {}

    for key in dict1.keys():
        new_dict[key] = (dict1[key], dict2[key])

    return new_dict


def probe_radius_to_debye_length(lambda_D, radius_probe):
    l = np.abs(lambda_D)  # Debye length [m]
    rp = np.abs(radius_probe)  # radius of probe [m]

    ratio = np.divide(rp, l)

    if ratio <= 3:
        sheath = "thick/OML"
    elif ratio >= 50:
        sheath = "thin"
    else:
        sheath = "transitional"

    other = {"sheath_type": sheath}
    return ratio, other


def find_intersection(coefs1, coefs2):
    m1, b1 = coefs1
    m2, b2 = coefs2
    # check if lines are parallel
    if m1 == m2:
        return None

    # calculate intersection point
    x_intersect = (b2 - b1) / (m1 - m2)
    y_intersect = m1 * x_intersect + b1

    return np.array([x_intersect, y_intersect])


def count_placeholders(fmt):
    count = 0
    L = string.Formatter().parse(fmt)
    for x in L:
        if x[1] is not None:
            count += 1
    return count


def get_rsquared(yexperimental, ypredicted):
    if len(yexperimental) != len(ypredicted):
        raise ValueError(
            "Vector 1 and Vector 2 arrays must have the same length.")

    corr_matrix = np.corrcoef(yexperimental, ypredicted)
    corr = corr_matrix[0, 1]
    rsq = corr**2

    return rsq


def determine_step(step: int | float, step_type: str, len_x: int) -> int:
    if step_type == 'index':
        istep = int(step)
    elif step_type == "percent":
        istep = int(step*len_x)
    elif step_type == "distance":
        raise NotImplementedError
    else:
        raise ValueError(
            "'step_type' does not match 'index' or 'percent': %s" % (str(step_type)))

    # make istep at least one
    if istep == 0:
        istep += 1
    return istep


def find_fit(
        xdata: npt.ArrayLike, ydata: npt.ArrayLike, deg: int = 1, power: int | float = 1, polarity: int = 1,
        reverse: bool = False, return_best: bool = False, fit_type: str = "polynomial",
        min_points: int = 5, istart: int | None = None, iend: int | None = None, invalid: str = "ignore",
        fstep: int | float | None = None, fstep_type: str | None = None, fstep_adjust: bool = True, fitmax: int | None = None,
        bstep: int | float | None = None, bstep_type: str | None = None, bstep_adjust: bool = True, bitmax: int | None = None,
        threshold_residual: int | float | None = None, threshold_rmse: int | float = 0.30,
        threshold_rsq: int | float = 0.95, threshold_method: str | None = None,
        convergence_residual_percent: int | float = 1.0, convergence_rmse_percent: int | float = 1.0,
        convergence_rsq_percent: int | float = 1.0, convergence_method: str | None = None,
        strict: bool = False, full: bool = False, printlog: bool = False) -> FunctionFit:
    # Convert xdata and ydata to numpy arrays
    xdata = np.array(xdata)
    ydata = np.array(ydata)

    # Set Defaults for threshold residual if not given
    threshold_residual = 0 if threshold_residual is None else threshold_residual

    # Set Defeaults for if reverse or standard direction
    if reverse:
        fstep = 0.1 if fstep is None else fstep
        fstep_type = "percent" if fstep_type is None else fstep_type
        bstep = 1 if bstep is None else bstep
        bstep_type = "index" if bstep_type is None else bstep_type
        fitmax = None if fitmax is None else (
            None if fitmax == "none" else fitmax)
        bitmax = None if bitmax is None else (  # changed 1 -> None
            None if bitmax == "none" else bitmax)
    else:
        bstep = 0.1 if bstep is None else bstep
        bstep_type = "percent" if bstep_type is None else bstep_type
        fstep = 1 if fstep is None else fstep
        fstep_type = "index" if fstep_type is None else fstep_type
        fitmax = None if fitmax is None else (  # changed 1 -> None
            None if fitmax == "none" else fitmax)
        bitmax = None if bitmax is None else (
            None if bitmax == "none" else bitmax)

    # verify keyword argument 'deg' is int
    verify_type(deg, int, 'deg')

    # verify keyword argument 'power' is int or float
    verify_type(power, (int, float), 'power')

    # verify keyword argument 'fit_type' is str
    verify_type(fit_type, str, 'fit_type')

    # correct deg if 'fit_type' == linear
    if fit_type == "linear":
        deg = 1

    # Set Defaults for threshold_method and convergence_method based on fit_type and deg
    if fit_type == "linear" or (fit_type == "polynomial" and deg == 1):
        threshold_method = "rsq" if threshold_method is None else threshold_method
        convergence_method = "rsq" if convergence_method is None else convergence_method
    else:
        threshold_method = "rmse" if threshold_method is None else threshold_method
        convergence_method = "rmse" if convergence_method is None else convergence_method

    # verify keyword arguments with bools are bool s.t. the loops below will not contiue if not True
    verify_type(reverse, bool, 'reverse')
    verify_type(return_best, bool, 'return_best')
    verify_type(fstep_adjust, bool, 'fstep_adjust')
    verify_type(bstep_adjust, bool, 'bstep_adjust')
    verify_type(strict, bool, 'strict')
    verify_type(full, bool, 'full')
    verify_type(printlog, bool, 'printlog')

    # Verify that these are arrays/iterable
    # and get lengths of data and check they are the same length
    verify_iterables(xdata, ydata)
    verify_1D_array(xdata)
    verify_1D_array(ydata)
    len_x = len(xdata)

    # convert convergence values from percent -> decimal for comparison later
    convergence_residual_percent_decimal = np.divide(
        convergence_residual_percent, 100)
    convergence_rmse_percent_decimal = np.divide(
        convergence_rmse_percent, 100)
    convergence_rsq_percent_decimal = np.divide(
        convergence_rsq_percent, 100)

    # convergence criteria
    if printlog:
        print("Fit Type: {}".format(fit_type))
        print("Strict Candidate: {0}".format(strict))
        print("Threshold Method: %s" % (threshold_method))
        print("Threshold Absolute Residual: {0:0.2e}".format(
            threshold_residual))
        print("Threshold RMSE %: {0:0.2%}".format(threshold_rmse))
        print("Threshold Rsquared: {0:.2f}".format(threshold_rsq))
        print("Convergence Method: %s" % (convergence_method))
        print("Convergence Residual Percent: {0:.2%}".format(
            convergence_residual_percent/100))
        print("Convergence RMSE Percent: {0:.2%}".format(
            convergence_rmse_percent/100))
        print("Convergence Rsquared Percent: {0:.2%}".format(
            convergence_rsq_percent/100))
        print(f"Reverse: {reverse}")
        print(f"Iteration Max: Forward: {fitmax}; Backward: {bitmax}")

    # initialize best trackers
    best_residual = float("inf")
    best_rmse = float("inf")
    best_rsq = float("-inf")
    best_fit = None

    # Initialize Exit, Threshold, and Convergence flags to finish
    threshold = False
    convergence = False
    exit = False

    # inintialize Return fit
    fit = None

    # initialize x,y arrays
    x = xdata[istart:iend]
    y = ydata[istart:iend]

    # if reverse direction
    if reverse:
        # initialize counter index backward
        jend = int(len_x) if iend is None else int(iend)
        # initiailize counter index forward
        jstart = 0 if istart is None else int(istart)
    else:
        # initialize counter index forward
        jstart = 0 if istart is None else int(istart)
        # initiailize counter index backward
        jend = int(len_x) if iend is None else int(iend)

    # initialize jfstep, jbstep index advancement counters
    jfstep = 0
    jbstep = 0

    # initialize inner and outer loop counters
    # initialize forward loop counter
    fiteration = 0
    # initialize backward loop counter
    biteration = 0

    # function for returning a fit class

    def create_fit():
        stats = {'rmse': rmse, 'residual': residual}
        stats.update(
            {'rsq': rsq}) if fit_type == "linear" or (fit_type == "polynomial" and deg == 1) else stats

        # if all components are wanted, added them to status dictionary
        stats.update({'rel_err': rel_diff_err,
                     'residuals': diff_err}) if full else stats

        # Create fit class depending on fit_type
        xrange = [np.min(x), np.max(x)]
        if fit_type == "linear" or fit_type == "polynomial":
            fit = PolynomialFit(poly, xrange, stats)
        elif fit_type == "exponential":
            fit = ExponentialFit(poly, xrange, stats)
        elif fit_type == "power":
            fit = PowerFit(power, poly, xrange, stats, polarity)
        else:
            raise ValueError("something went wrong in creating fit class")
        return fit

    # outer loop helper function
    def outer():
        out = ((jend >= min_points if istart is None else istart+min_points) if reverse else (
            jstart <= len_x-min_points if iend is None else iend-min_points)) and (
            (biteration < bitmax if bitmax is not None else True) if reverse else
            (fiteration < fitmax if fitmax is not None else True))
        return out

    # inner loop helper function
    def inner():
        out = (jstart <= jend-min_points) if reverse else (jend >= jstart+min_points) and (
            (fiteration < fitmax if fitmax is not None else True) if reverse else
            (biteration < bitmax if bitmax is not None else True))

        return out

    # Start outer loop to continue till the length of array from start index to end is less than minimum points
    while outer():

        # Restart residual and rsme convergence trackers because the outerloop restarted and would not be a fair comparison
        last_residual = float("inf")
        last_rmse = float("inf")
        last_rsq = float("inf")

        # Start inner loop to continue till the length of the array between start index and to end index is less than minimum points
        while inner():
            # Grab the arrays with specified start indexes
            x = xdata[jstart:jend]
            y = ydata[jstart:jend] * polarity

            if fit_type == "linear" or (fit_type == "polynomial" and deg == 1):
                # Determine the Coefs for linear (same as polynomial but deg is automatically set to 1 no matter what
                # This uses  ~numpy.polonomial.Polonomial.fit (note: this is least squares fit )
                poly = P.fit(x, y, deg=deg)
                # y-values on standared plot using x-values of actual data area of interest
                yp = poly.convert()(x)

            elif fit_type == "polynomial":
                # Determine the Coefs for polynomial using regression fit using ~numpy.polonomial.Polonomial.fit (note: this is least squares fit )
                poly = P.fit(x, y, deg=deg)
                # y-values on standared plot using x-values of actual data area of interest
                yp = poly.convert()(x)

            elif fit_type == "exponential":
                # Take natural log of data
                logy = np.log(y)
                # Determine the Coefs for semilogy using linear regression fit in semilogy space (note: this is non-linear regression)
                poly = P.fit(x, logy, 1, w=np.sqrt(y))
                # poly = P.fit(x, logy, 1)
                # popt, opt = curve_fit(_fit_exponential_func, x,
                #                      y, p0=poly.convert().coef)
                # poly.coef = np.polynomial.polyutils.mapdomain(
                #    popt, [-1, 1], poly.domain)
                # poly.domain = [-1, 1]

                # y-values on standared plot using x-values of actual data area of interest
                yp = np.exp(poly.convert()(x))

            elif fit_type == "power":
                # Take inverse power of data
                mody = np.power(y, 1/power)
                # Determine the Coefs for fit using linear regression fit in mod power space (note: this is non-linear regression)
                poly = P.fit(x, mody, 1)

                # y-values on standared plot using x-values of actual data area of interest
                temp = poly.convert()(x)
                if invalid == "ignore":
                    with np.errstate(invalid='ignore'):
                        yp = np.power(temp, power)
                else:
                    yp = np.power(temp, power)

                # correct for polarity
                # yp *= polarity

                # check for nan because of the root
                inan = np.argwhere(np.isnan(yp))

                # remove nans for comparisons
                # y = np.delete(y, inan)
                # yp = np.delete(yp, inan)

            else:
                raise ValueError("'fit_type' is not a known value of 'linear', 'polynomial', 'exponential', or 'power': {}".format(
                    fit_type
                ))

            # The following are used to determine if the fit is good
            # Two criteria maybe used to determine the goodness of fit
            # Root Mean Square Error (RMSE) of the Relative Difference Error -> rmse_rel_diff_err
            # Absolute Residual (Residual) - The sum of the absolute difference between input (experimental) and the fitted (theoretical) -> residual
            # Difference Error -> diff_err
            diff_err = np.subtract(y, yp)
            # Absolue Difference Error -> abs_diff_err
            abs_diff_err = np.abs(diff_err)
            # Relative Difference Percent Error (decimal form) -> rel_diff_err
            rel_diff_err = np.divide(diff_err, yp)
            # Relative Difference Percent Error (percent from) -> rel_per_diff_err (not used currently, maybe removed)
            rel_per_diff_err = np.multiply(rel_diff_err, 100)
            # Root Mean Square of Relative Differnece Percent Error (decimal form) -> rmse_rel_diff_err
            rmse_rel_diff_err = np.sqrt(np.mean(np.power(rel_diff_err, 2)))
            # Root Mean Square of Relative Differnece Percent Error (percent form) -> rmse_rel_per_diff_err (not used currently, maybe removed)
            rmse_rel_per_diff_err = np.multiply(rmse_rel_diff_err, 100)

            # rmse value used for comparison (convience only)
            rmse = rmse_rel_diff_err.copy()

            # total absolute residual -> residual
            residual = np.sum(abs_diff_err)

            # correlation coefs for Rsquared (not really used outside of 'linear' fit_type)
            corr_matrix = np.corrcoef(y, yp)
            corr = corr_matrix[0, 1]
            rsq = corr*corr

            # Calculate convergence for both Residual and RMSE
            # get the decimal form of relative difference of residual from last point
            convergence_residual = np.abs(np.divide(np.subtract(
                last_residual, residual), last_residual)) if last_residual != float("inf") else float("inf")
            # get the difference of in decimal form of the last two rmse
            convergence_rmse = np.abs(np.subtract(
                last_rmse, rmse)) if last_rmse != float("inf") else float("inf")
            # get the difference of in decimal form of the last two Rsquared
            convergence_rsq = np.abs(np.subtract(
                last_rsq, rsq)) if last_rsq != float("-inf") else float("-inf")

            # After last point was a possible candidate for a solution, check for convergence by checking this point to see the magnitude of the change in solution
            # test this point for convergence, if converges, breaks here so last point is returned and this point is disregared as criteria were meet on last point
            # the input arg 'convergence_method' determines if residual or rmse are used for determining convergence
            if threshold:
                if convergence_method == "both":
                    convergence = convergence_residual <= convergence_residual_percent_decimal and convergence_rmse <= convergence_rmse_percent_decimal
                elif convergence_method == "rmse":
                    convergence = convergence_rmse <= convergence_rmse_percent_decimal
                elif convergence_method == "threshold_residual":
                    convergence = convergence_residual <= convergence_residual_percent_decimal
                elif convergence_method == "rsq" or convergence_method == "rsquared":
                    convergence = convergence_rsq <= convergence_rsq_percent_decimal
                else:
                    raise ValueError("'convergence_method' can only be 'both', 'rmse', 'residual' or 'rsq': %s" % (
                        str(convergence_method)))

            # check if residual is less than threshold and RMSE is less
            if threshold_method == "both":
                threshold = residual <= threshold_residual and rmse <= threshold_rmse
            elif threshold_method == "rmse":
                threshold = rmse_rel_diff_err <= threshold_rmse
            elif threshold_method == "residual":
                threshold = residual <= threshold_residual
            elif threshold_method == "rsq" or threshold_method == "rsquared":
                threshold = rsq >= threshold_rsq
            else:
                raise ValueError("'threshold_method' can only be 'both', 'rmse', 'residual', or 'rsq': %s" % (
                    str(threshold_method)))

            # If strict, threshold and convergence must both be meet for both canidite point and convergence test point
            # else, only convergence has to be made between the two points since threshold was meet on the canidite point
            # Then, exit is set to true to break
            exit = threshold and convergence if strict else convergence

            # If printlog is True, show steps in printed log
            if printlog:
                print("----------------------------------------------")
                print("jstart: {0}\tjend: {1}\tforward step: {2}\tbackward step: {3}".format(
                    jstart, jend, jfstep, jbstep))
                print("Absolute Residual: {0:.2e}\tBest Abs. Residual: {1:.2e}\tLast Abs. Residual: {2:.2e}".format(
                    residual, best_residual, last_residual))
                print("RMSE: {0:.2%}\tBest RMSE: {1:.2%}\tLast RMSE: {2:.2%}".format(
                    rmse_rel_diff_err, best_rmse, last_rmse))
                print("Relative Difference: Max Rel Diff: {0:.2%};\tMin Rel Diff: {1:.2%}".format(
                    np.max(rel_diff_err), np.min(rel_diff_err)))
                print("Rsquared: {0:.2f}\tBest Rsquared: {1:.2f}\tLast Rsquared: {2:.2f}".format(
                    rsq, best_rsq, last_rsq))
                print("Convergence Abs. Residual: {0:.2%}\tConvergence RMSE: {1:.2%}\tConvergence Rsquared: {2:.2%}".format(
                    convergence_residual, convergence_rmse, convergence_rsq))
                print("Candidate:\tThreshold: {0}\tConvergence: {1}\nExit: {2}".format(
                    threshold, convergence, exit))
                # if exit, should quit and add this extra line below
                if exit:
                    print("----------------------------------------------")

            # Exit Loop
            if exit:
                break

            # Have not meet exit criteria, keep updating last interation data if candidate (threshold==True)
            if threshold:
                # Returned Dictionary depending if input keyword arg is True
                # These values are good as they were saved before the next loop, therefore they will only besaved for candidate
                # adds rsq if linear
                fit = create_fit()

            # if the best fit option is given, the best fit is also stored temporarly and returned if convergence is not made
            if return_best:
                if threshold:
                    best_fit = fit
                else:

                    if threshold_method == "both" or threshold_method == "residual":
                        # Keep track of best residual (not neccessarily the best fit)
                        if residual <= best_residual:
                            best_residual = residual.copy()
                            best_fit = create_fit()

                    elif threshold_method == "both" or threshold_method == "rmse":
                        # Keep track of best RMSE (not neccessarily the best fit)
                        if rmse <= best_rmse:
                            best_rmse = rmse.copy()
                            best_fit = create_fit()

                    elif threshold_method == "rsq" or threshold_method == "rsquared":
                        # Keep track of best Rsquared (not neccessarily the best fit)
                        if rsq >= best_rsq:
                            best_rsq = rsq.copy()
                            best_fit = create_fit()

                    else:
                        raise ValueError("something went wrong in return_best")

            # Update the last residual and rmse (must be after printlog)
            last_residual = residual.copy()
            last_rmse = rmse.copy()
            last_rsq = rsq.copy()

            # If reverse, move forward, else move backward
            if reverse:
                # If break does not happen yet, determine forward step size
                jfstep = determine_step(fstep, fstep_type, len(
                    x)) if fstep_adjust else determine_step(fstep, fstep_type, len_x)
                # Forward counter index is advanced with step
                jstart += jfstep
                # advance interation
                fiteration += 1
            else:
                # If the break did not happen yet, determine the backward step size
                jbstep = determine_step(bstep, bstep_type, len(
                    x)) if bstep_adjust else determine_step(bstep, bstep_type, len_x)
                # the backward counter index is advanced with step
                jend -= jbstep
                # advance interation
                biteration += 1

        # Exits the outerloop if criteria meet
        if exit:
            break

        # Restart conter index for inner loop
        if reverse:
            # initiailize or restart counter index forward
            jstart = 0 if istart is None else int(istart)
            # restart inner loop couter
            fiteration = 0
        else:
            # initiailize or restart counter index backward
            jend = int(len_x) if iend is None else int(iend)
            # restart inner loop couter
            biteration = 0

        # For Outer loop: If reverse, move backward, else move forward
        if reverse:
            # If the break did not happen yet, determine the backward step size
            jbstep = determine_step(bstep, bstep_type, len(
                x)) if bstep_adjust else determine_step(bstep, bstep_type, len_x)
            # the backward counter index is advanced with step
            jend -= jbstep
            # advance outer interation
            biteration += 1
        else:
            # If break does not happen yet, determine forward step size
            jfstep = determine_step(fstep, fstep_type, len(
                x)) if fstep_adjust else determine_step(fstep, fstep_type, len_x)
            # Forward counter index is advanced with step
            jstart += jfstep
            # advance outer interation
            fiteration += 1

    # if exit is True, then criteria meet and can exit
    if exit:
        if fit:
            return fit
        else:
            raise MissingReturnError
    elif return_best:
        if best_fit:
            return best_fit
        else:
            raise FitConvergenceError
    else:
        raise FitConvergenceError


def normalize(x: np.ndarray) -> np.ndarray:
    """
    Returns the normalized ndarray of input array

    Parameters
    ----------
    x : arraylike
        A 1D array to be normalized.

    Returns
    -------
    np.ndarray
        Normalized np.ndarray.
    """
    return (x-np.min(x))/(np.max(x)-np.min(x))


def residual(xdata: np.ndarray, xdata_mean: np.ndarray):
    """
    Returns the residual (difference) between the two arrays

    Parameters
    ----------
    xdata :  arraylike
        A 1D array of experimental data
    xdata_mean : arraylike
        A 1D array of mean of the experimental data

    Returns
    -------
    np.ndarray
        Residual
    """
    # verify same length


def preproces_data(xdata: np.ndarray) -> np.ndarray:

    # initialize and set default
    outdata = np.ndarray(np.size(xdata), dtype=None)
    return outdata


def _fit_linear_func(x, b, m):
    return m*x+b


def _fit_exponential_func(x, *args):
    total = _fit_polynomial_func(x, *args)
    return np.exp(total)


def _fit_power_func(power, x, *args, invalid="ignore"):
    total = _fit_polynomial_func(x, *args)
    # print(total) if np.any(np.isnan(total)) else None
    # print(power)
    if invalid == "ignore":
        with np.errstate(invalid="ignore"):
            res = np.power(total, power)
    else:
        res = np.power(total, power)
    # print(res) if np.any(np.isnan(res)) else None
    return res


def _fit_polynomial_func(x, *args):
    total = 0.0
    for k, arg in enumerate(args):
        total = np.add(
            total,
            np.multiply(
                arg,
                np.power(
                    x,
                    k
                )
            )
        )

    return total



import matplotlib.pyplot as plt
import matplotlib.animation as animation
import time
import datetime as dt
from pstl.instruments.daq.agilent.gpib import Agilent34970A as DAQ

class ALARM():
    def __init__(self,upper=130,lower=105):
        self.upper=upper
        self.lower=lower

class SUBPLOT():
    def __init__(self,ax,location=None):
        self.ax=ax
        self.location=location
        self.x=[]
        self.y=[]

        self.ylim=[95,135]
        self.alarm=ALARM()


class MONITOR():
    def __init__(self,nrows,ncols):
        self.nrows=nrows
        self.ncols=ncols

        fig,ax_list=plt.subplots(nrows=3,ncols=3)
        self.figure=fig

        ax=ax_list.ravel()
        self.axes=ax

        nax=len(ax)
        self.nax=nax
        subplot=[None]*nax
        r=[None]*nax
        for k in range(nax):
            subplot[k]=SUBPLOT(ax[k],int(112+k))
        self.subplot=subplot

        #daq=DAQ("GPIB0::10::INSTR")
        daq=DAQ()

        daq.addCardAgilent34901A(1,20,'TCK')


        self.daq=daq

    # This function is called periodically from FuncAnimation
    def animate(self, i):

        for k in range(self.nax):
            # handel
            dis=self.subplot[k]

            # set xs,ys
            xs=self.subplot[k].x
            ys=self.subplot[k].y

            # axes to use
            ax=self.axes[k]

            # Read temperature 
            temp_c = round(float(self.daq.get(self.subplot[k].location)),2)

            # Add x and y to lists
            xs.append(dt.datetime.now().strftime('%H:%M:%S.%f'))
            ys.append(temp_c)

            # Limit x and y lists to 20 items
            xs = xs[-20:]
            ys = ys[-20:]

            # Draw x and y lists
            ax.clear()
            ax.plot(xs, ys,\
                marker='o',linestyle='-',markerfacecolor='none')

            # horizontal limit alarm
            ax.axhline(y=dis.alarm.upper,color='r')
            ax.axhline(y=dis.alarm.lower,color='r')

            self.subplot[k].x=xs
            self.subplot[k].y=ys

            # Format plot
            #ax.set_xticks(rotation=45, ha='right')
            ax.grid(True)
            ax.set_xticklabels([])
            #ax.set_xticks([])
            if ys[-1]>=dis.ylim[0] and ys[-1]<=dis.ylim[1]:
                ax.set(ylim=dis.ylim)
            #plt.subplots_adjust(bottom=0.30)
            ax.set(title='TEMP at %s'%(str(dis.location)))
            ax.set(xlabel='Time [s]')
            ax.set(ylabel='Temperature [deg C]')

    def monitor(self):
        
        # Set up plot to call animate() function periodically
        ani = animation.FuncAnimation(self.figure, self.animate, fargs=None, interval=1000)
        wm = plt.get_current_fig_manager()
        wm.window.state('zoomed')
        plt.show()


def monitor():
    temperatures=MONITOR(3,3)
    print("monitoring...")
    temperatures.monitor()






from matplotlib.widgets import Cursor
import numpy as np

from matplotlib.backend_bases import MouseEvent


class AnnotatedCursor(Cursor):
    """
    A crosshair cursor like `~matplotlib.widgets.Cursor` with a text showing \
    the current coordinates.

    For the cursor to remain responsive you must keep a reference to it.
    The data of the axis specified as *dataaxis* must be in ascending
    order. Otherwise, the `numpy.searchsorted` call might fail and the text
    disappears. You can satisfy the requirement by sorting the data you plot.
    Usually the data is already sorted (if it was created e.g. using
    `numpy.linspace`), but e.g. scatter plots might cause this problem.
    The cursor sticks to the plotted line.

    Parameters
    ----------
    line : `matplotlib.lines.Line2D`
        The plot line from which the data coordinates are displayed.

    numberformat : `python format string <https://docs.python.org/3/\
    library/string.html#formatstrings>`_, optional, default: "{0:.4g};{1:.4g}"
        The displayed text is created by calling *format()* on this string
        with the two coordinates.

    offset : (float, float) default: (5, 5)
        The offset in display (pixel) coordinates of the text position
        relative to the cross-hair.

    dataaxis : {"x", "y"}, optional, default: "x"
        If "x" is specified, the vertical cursor line sticks to the mouse
        pointer. The horizontal cursor line sticks to *line*
        at that x value. The text shows the data coordinates of *line*
        at the pointed x value. If you specify "y", it works in the opposite
        manner. But: For the "y" value, where the mouse points to, there might
        be multiple matching x values, if the plotted function is not biunique.
        Cursor and text coordinate will always refer to only one x value.
        So if you use the parameter value "y", ensure that your function is
        biunique.

    Other Parameters
    ----------------
    textprops : `matplotlib.text` properties as dictionary
        Specifies the appearance of the rendered text object.

    **cursorargs : `matplotlib.widgets.Cursor` properties
        Arguments passed to the internal `~matplotlib.widgets.Cursor` instance.
        The `matplotlib.axes.Axes` argument is mandatory! The parameter
        *useblit* can be set to *True* in order to achieve faster rendering.

    """

    def __init__(self, line, numberformat="{0:.4g};{1:.4g}", offset=(5, 5),
                 dataaxis='x', textprops=None, **cursorargs):
        if textprops is None:
            textprops = {}
        # The line object, for which the coordinates are displayed
        self.line = line
        # The format string, on which .format() is called for creating the text
        self.numberformat = numberformat
        # Text position offset
        self.offset = np.array(offset)
        # The axis in which the cursor position is looked up
        self.dataaxis = dataaxis

        # First call baseclass constructor.
        # Draws cursor and remembers background for blitting.
        # Saves ax as class attribute.
        Cursor.__init__(self, **cursorargs)

        # Default value for position of text.
        self.set_position(self.line.get_xdata()[0], self.line.get_ydata()[0])
        # Create invisible animated text
        self.text = self.ax.text(
            self.ax.get_xbound()[0],
            self.ax.get_ybound()[0],
            "0, 0",
            animated=bool(self.useblit),
            visible=False, **textprops)
        # The position at which the cursor was last drawn
        self.lastdrawnplotpoint = None

    def onmove(self, event):
        """
        Overridden draw callback for cursor. Called when moving the mouse.
        """

        # Leave method under the same conditions as in overridden method
        if self.ignore(event):
            self.lastdrawnplotpoint = None
            return
        if not self.canvas.widgetlock.available(self):
            self.lastdrawnplotpoint = None
            return

        # If the mouse left drawable area, we now make the text invisible.
        # Baseclass will redraw complete canvas after, which makes both text
        # and cursor disappear.
        if event.inaxes != self.ax:
            self.lastdrawnplotpoint = None
            self.text.set_visible(False)
            Cursor.onmove(self, event)
            return

        # Get the coordinates, which should be displayed as text,
        # if the event coordinates are valid.
        plotpoint = None
        if event.xdata is not None and event.ydata is not None:
            # Get plot point related to current x position.
            # These coordinates are displayed in text.
            plotpoint = self.set_position(event.xdata, event.ydata)
            # Modify event, such that the cursor is displayed on the
            # plotted line, not at the mouse pointer,
            # if the returned plot point is valid
            if plotpoint is not None:
                event.xdata = plotpoint[0]
                event.ydata = plotpoint[1]

        # If the plotpoint is given, compare to last drawn plotpoint and
        # return if they are the same.
        # Skip even the call of the base class, because this would restore the
        # background, draw the cursor lines and would leave us the job to
        # re-draw the text.
        if plotpoint is not None and plotpoint == self.lastdrawnplotpoint:
            return

        # Baseclass redraws canvas and cursor. Due to blitting,
        # the added text is removed in this call, because the
        # background is redrawn.
        super().onmove(event)

        # Check if the display of text is still necessary.
        # If not, just return.
        # This behaviour is also cloned from the base class.
        if not self.get_active() or not self.visible:
            return

        # Draw the widget, if event coordinates are valid.
        if plotpoint is not None:
            # Update position and displayed text.
            # Position: Where the event occurred.
            # Text: Determined by set_position() method earlier
            # Position is transformed to pixel coordinates,
            # an offset is added there and this is transformed back.
            temp = [event.xdata, event.ydata]
            temp = self.ax.transData.transform(temp)
            temp = temp + self.offset
            temp = self.ax.transData.inverted().transform(temp)
            self.text.set_position(temp)
            self.text.set_text(self.numberformat.format(*plotpoint))
            self.text.set_visible(self.visible)

            # Tell base class, that we have drawn something.
            # Baseclass needs to know, that it needs to restore a clean
            # background, if the cursor leaves our figure context.
            self.needclear = True

            # Remember the recently drawn cursor position, so events for the
            # same position (mouse moves slightly between two plot points)
            # can be skipped
            self.lastdrawnplotpoint = plotpoint
        # otherwise, make text invisible
        else:
            self.text.set_visible(False)

        # Draw changes. Cannot use _update method of baseclass,
        # because it would first restore the background, which
        # is done already and is not necessary.
        if self.useblit:
            self.ax.draw_artist(self.text)
            self.canvas.blit(self.ax.bbox)
        else:
            # If blitting is deactivated, the overridden _update call made
            # by the base class immediately returned.
            # We still have to draw the changes.
            self.canvas.draw_idle()

    def set_position(self, xpos, ypos):
        """
        Finds the coordinates, which have to be shown in text.

        The behaviour depends on the *dataaxis* attribute. Function looks
        up the matching plot coordinate for the given mouse position.

        Parameters
        ----------
        xpos : float
            The current x position of the cursor in data coordinates.
            Important if *dataaxis* is set to 'x'.
        ypos : float
            The current y position of the cursor in data coordinates.
            Important if *dataaxis* is set to 'y'.

        Returns
        -------
        ret : {2D array-like, None}
            The coordinates which should be displayed.
            *None* is the fallback value.
        """

        # Get plot line data
        xdata = self.line.get_xdata()
        ydata = self.line.get_ydata()

        # The dataaxis attribute decides, in which axis we look up which cursor
        # coordinate.
        if self.dataaxis == 'x':
            pos = xpos
            data = xdata
            lim = self.ax.get_xlim()
        elif self.dataaxis == 'y':
            pos = ypos
            data = ydata
            lim = self.ax.get_ylim()
        else:
            raise ValueError(f"The data axis specifier {self.dataaxis} should "
                             f"be 'x' or 'y'")

        # If position is valid and in valid plot data range.
        if pos is not None and lim[0] <= pos <= lim[-1]:
            # Find closest x value in sorted x vector.
            # This requires the plotted data to be sorted.
            index = np.searchsorted(data, pos)
            # Return none, if this index is out of range.
            if index < 0 or index >= len(data):
                return None
            # Return plot point as tuple.
            return (xdata[index], ydata[index])

        # Return none if there is no good related point for this x position.
        return None

    def clear(self, event):
        """
        Overridden clear callback for cursor, called before drawing the figure.
        """

        # The base class saves the clean background for blitting.
        # Text and cursor are invisible,
        # until the first mouse move event occurs.
        Cursor.clear(self, event)
        if self.ignore(event):
            return
        self.text.set_visible(False)

    def _update(self):
        """
        Overridden method for either blitting or drawing the widget canvas.

        Passes call to base class if blitting is activated, only.
        In other cases, one draw_idle call is enough, which is placed
        explicitly in this class (see *onmove()*).
        In that case, `~matplotlib.widgets.Cursor` is not supposed to draw
        something using this method.
        """

        if self.useblit:
            Cursor._update(self)  # type: ignore



from abc import ABC, abstractmethod, abstractproperty

import numpy as np

from pstl.diagnostics.probes.classes import Probe

available_plasma_properties = [
    "V_f",
    "V_s",
    "KT_e",
    "lambda_De",
    "n_e",
    "n_i",
    "I_es",
    "I_is",
    "J_es",
    "J_is",
    "sheath",
]


class SingleProbeLangmuir(Probe, ABC):
    def __init__(self, diameter, *args, shape="Unknown", **kwargs) -> None:
        self._diameter = float(diameter)
        self._radius = diameter/2
        self._shape = shape

        self._area = self.calc_area(diameter, *args, **kwargs)

    @property
    def diameter(self):
        return self._diameter

    @property
    def radius(self):
        return self._radius

    @property
    def area(self):
        return self._area

    @property
    def shape(self):
        return self._shape

    @abstractmethod
    def calc_area(self, diameter, *args, **kwargs) -> float:
        pass


class CylindericalSingleProbeLangmuir(SingleProbeLangmuir):
    def __init__(self, diameter, length, *args, **kwargs) -> None:
        shape = "cylinderical"
        super().__init__(diameter, length, *args, shape=shape, **kwargs)

        self._length = float(length)

    @property
    def length(self):
        return self._length

    def calc_area(self, diameter, length, *args, **kwargs) -> float:
        super().calc_area(diameter, length, *args, **kwargs)
        return diameter*np.pi*(length + diameter/4)


class PlanarSingleProbeLangmuir(SingleProbeLangmuir):
    def __init__(self, diameter, *args, **kwargs) -> None:
        shape = "planar"
        super().__init__(diameter, *args, shape=shape, **kwargs)

    def calc_area(self, diameter, *args, **kwargs) -> float:
        super().calc_area(diameter, *args, **kwargs)
        return diameter*diameter*np.pi/4


class SphericalSingleProbeLangmuir(SingleProbeLangmuir):
    def __init__(self, diameter, length=None, *args, **kwargs) -> None:
        if length is None:
            length = diameter

        shape = "spherical"
        super().__init__(diameter, length, *args, shape=shape, **kwargs)

        self._length = float(length)

    @property
    def length(self):
        return self._length

    def calc_area(self, diameter, length, *args, **kwargs) -> float:
        super().calc_area(diameter, length, *args, **kwargs)
        height = length-diameter/2
        radius = diameter/2
        cap = 2*np.pi*(radius*height)
        hemisphere = 2*np.pi*radius*radius
        return cap+hemisphere



def convert2hex(num,maximum,hexalen):
    """Converts decimal integer to hexadecimal and returns the last three
    numbers/letters in uppercase to be passed to kepco_BOP

    num and maximum must be int or floats

    returns a string of length 3 of corresponding hexadecimal format

    FFF/X = 4095/E = Emax/E
    where X is your desired value in hex
    FFF is max hex (4095)
    E is the desired value in decimal
    Emax is the max value in decimal

    X=INT(FFF*(E/Emax))

    Note: 4095 (decimal) == FFF (hex)
    """
    if not isinstance(num,(int,float)):
        error=("'num' variable '%s' is not an integer or float"%(str(num)))
        raise TypeError(error)
    if not isinstance(maximum,(int,float)):
        error=("'maximum' variable '%s' is not an integer or float"%(str(maximum)))
        raise TypeError(error)
    if not isinstance(hexalen,int):
        error=("'hexalen' variable '%s' is not an integer"%(str(hexalen)))
        raise TypeError(error)
    if num==0:
        hexa="0"
    else:
        hexamax=int(pow(16,hexalen)-1) 
        hexa=hex(int(hexamax/(maximum/num)))[2:].upper()
    hexa=extendHexa(hexa,hexalen)
    return hexa

def extendHexa(hexa,strlen):
    hexalen=len(hexa)
    while hexalen<strlen:
        hexa="0"+hexa
        hexalen=len(hexa)
    return hexa

def polarity(string):
    try:    # test if the input is string or number
        val=int(string)
    except: # not number, see if string
        if isinstance(string,str):
            if string=="+" or string=="positive" or string=="pos":
                val=1
            elif string=="-" or string=="negative" or string=="neg":
                val=-1
            else: # produces error
                val=string
        else:
            raise TypeError("Input was not of type int,float, or str"+\
                    "But of type %s"%(str(type(string))))
    finally:
        if val==-1:
            out=val
        elif val==1:
            out=val
        else:
            raise TypeError("'%s' cannot correlate to -1 or 1"%(str(string)))
    return out






import inspect

import numpy as np


def add_empty_dict(func):
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result, {}
    return wrapper


def absorb_extra_args_kwargs_old(func):
    def wrapper(*args, **kwargs):
        print("Before the function is executed.")
        # absorb any extra arguments and keyword arguments here
        extra_args = [
            arg for arg in args if arg not in func.__code__.co_varnames]
        extra_kwargs = {key: val for key, val in kwargs.items(
        ) if key not in func.__code__.co_varnames}
        result = func(*args, **kwargs)
        print("After the function is executed.")
        return result
    return wrapper


def absorb_extra_args_kwargs(func):
    def wrapper(*args, **kwargs):
        # Separate the defined args and kwargs from the extra ones
        defined_args = args[:func.__code__.co_argcount]
        defined_kwargs = {
            k: v for k, v in kwargs.items() if k in func.__code__.co_varnames}

        extra_args = args[func.__code__.co_argcount:]
        extra_kwargs = {k: v for k, v in kwargs.items(
        ) if k not in func.__code__.co_varnames}

        # Call the decorated function with the defined args and kwargs
        result = func(*defined_args, **defined_kwargs)

        # Output a message indicating that the function has been executed
        # print(f"{func.__name__} has been executed")

        return result

    return wrapper


def where_function_else_zero(func, condition_func):

    def wrapper(*args, **kwargs):
        condition = condition_func(*args, **kwargs)
        return np.where(condition, func(*args, **kwargs), 0)
    return wrapper


def dict_args(func):
    sig = inspect.signature(func)
    arg_names = tuple(sig.parameters.keys())

    def wrapper(args_dict):
        args = tuple(args_dict.get(name, inspect.Parameter.empty)
                     for name in arg_names)
        return func(*args)

    return wrapper
